# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from ax.core import MultiObjectiveOptimizationConfig

from ax.core.auxiliary import AuxiliaryExperiment, AuxiliaryExperimentPurpose
from ax.core.experiment import Experiment

from ax.core.trial_status import TrialStatus
from ax.exceptions.core import DataRequiredError, UserInputError
from ax.exceptions.generation_strategy import MaxParallelismReachedException

if TYPE_CHECKING:
    from ax.generation_strategy.generation_node import GenerationNode

from ax.utils.common.base import SortableBase
from ax.utils.common.serialization import SerializationMixin, serialize_init_args
from pyre_extensions import none_throws


DATA_REQUIRED_MSG = (
    "All trials for current node {node_name} have been generated, "
    "but not enough data has been observed to proceed to the next "
    "Generation node. Try again when more is are available."
)


class TransitionCriterion(SortableBase, SerializationMixin):
    """
    Simple class to describe a condition which must be met for this GenerationNode to
    take an action such as generation, transition, etc.

    Args:
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to when this criterion is met, if it exists.
        block_gen_if_met: A flag to prevent continued generation from the
            associated GenerationNode if this criterion is met but other criterion
            remain unmet. Ex: ``MinTrials`` has not been met yet, but
            MinTrials has been reached. If this flag is set to true on MinTrials then
            we will raise an error, otherwise we will continue to generate trials
            until ``MinTrials`` is met (thus overriding MinTrials).
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: MaxGenerationParallelism
            defaults to setting this to False since we can complete and move on from
            this node without ever reaching its threshold.
        continue_trial_generation: A flag to indicate that all generation for a given
            trial is not completed, and thus even after transition, the next node will
            continue to generate arms for the same trial. Example usage: in
            ``BatchTrial``s we may  enable generation of arms within a batch from
            different ``GenerationNodes`` by setting this flag to True.
    """

    _transition_to: str | None = None

    def __init__(
        self,
        transition_to: str | None = None,
        block_transition_if_unmet: bool | None = True,
        block_gen_if_met: bool | None = False,
        continue_trial_generation: bool | None = False,
    ) -> None:
        self._transition_to = transition_to
        self.block_transition_if_unmet = block_transition_if_unmet
        self.block_gen_if_met = block_gen_if_met
        self.continue_trial_generation = continue_trial_generation

    @property
    def transition_to(self) -> str | None:
        """The name of the next GenerationNode after this TransitionCriterion is
        completed, if it exists.
        """
        return self._transition_to

    @abstractmethod
    def is_met(
        self,
        experiment: Experiment,
        curr_node: GenerationNode,
    ) -> bool:
        """If the criterion of this TransitionCriterion is met, returns True."""
        pass

    @abstractmethod
    def block_continued_generation_error(
        self,
        node_name: str,
        experiment: Experiment,
        trials_from_node: set[int],
    ) -> None:
        """Error to be raised if the `block_gen_if_met` flag is set to True."""
        pass

    @property
    def criterion_class(self) -> str:
        """Name of the class of this TransitionCriterion."""
        return self.__class__.__name__

    def __repr__(self) -> str:
        return f"{self.criterion_class}({serialize_init_args(obj=self)})"

    @property
    def _unique_id(self) -> str:
        """Unique id for this TransitionCriterion."""
        # TODO @mgarrard validate that this is unique enough
        return str(self)


class AutoTransitionAfterGen(TransitionCriterion):
    """A class to designate automatic transition from one GenerationNode to another.

    Args:
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to next.
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: This criterion defaults to
            setting this to True to ensure we validate a GeneratorRun is generated by
            the current GenerationNode.
        continue_trial_generation: A flag to indicate that all generation for a given
            trial is not completed, and thus even after transition, the next node will
            continue to generate arms for the same trial. Example usage: in
            ``BatchTrial``s we may  enable generation of arms within a batch from
            different ``GenerationNodes`` by setting this flag to True.
    """

    def __init__(
        self,
        transition_to: str,
        block_transition_if_unmet: bool | None = True,
        continue_trial_generation: bool | None = True,
    ) -> None:
        super().__init__(
            transition_to=transition_to,
            block_transition_if_unmet=block_transition_if_unmet,
            continue_trial_generation=continue_trial_generation,
        )

    def is_met(
        self,
        experiment: Experiment,
        curr_node: GenerationNode,
    ) -> bool:
        """Return True as soon as any GeneratorRun is generated by this
        GenerationNode.
        """
        # Handle edge case where the InputConstructor for a GenerationNode
        # with this criterion requests no arms to be generated, therefore, indicating
        # that this GenerationNode should be skipped and so we can transition to the
        # next node as defined by this criterion.
        if curr_node._should_skip:
            return True
        last_gr_from_gs = curr_node.generation_strategy.last_generator_run
        return (
            last_gr_from_gs._generation_node_name == curr_node.node_name
            if last_gr_from_gs is not None
            else False
        )

    def block_continued_generation_error(
        self,
        node_name: str,
        experiment: Experiment,
        trials_from_node: set[int],
    ) -> None:
        """Error to be raised if the `block_gen_if_met` flag is set to True."""
        pass


class IsSingleObjective(TransitionCriterion):
    """A class to initiate transition based on whether the experiment is optimizing
    for a single objective or multiple objectives.

    Args:
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to next.
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: This criterion defaults to
            setting this to True to ensure we validate a GeneratorRun is generated by
            the current GenerationNode.
        continue_trial_generation: A flag to indicate that all generation for a given
            trial is not completed, and thus even after transition, the next node will
            continue to generate arms for the same trial. Example usage: in
            ``BatchTrial``s we may  enable generation of arms within a batch from
            different ``GenerationNodes`` by setting this flag to True.
    """

    def __init__(
        self,
        transition_to: str,
        block_transition_if_unmet: bool | None = True,
        continue_trial_generation: bool | None = False,
    ) -> None:
        super().__init__(
            transition_to=transition_to,
            block_transition_if_unmet=block_transition_if_unmet,
            continue_trial_generation=continue_trial_generation,
        )

    def is_met(
        self,
        experiment: Experiment,
        curr_node: GenerationNode,
    ) -> bool:
        """Return True if the optimization config is not of type
        ``MultiObjectiveOptimizationConfig``."""
        return (
            not isinstance(
                experiment.optimization_config, MultiObjectiveOptimizationConfig
            )
            if experiment.optimization_config is not None
            else True
        )

    def block_continued_generation_error(
        self,
        node_name: str,
        experiment: Experiment,
        trials_from_node: set[int],
    ) -> None:
        # TODO: @mgarrard add error message that makes sense
        pass


class TrialBasedCriterion(TransitionCriterion):
    """Common class for transition criterion that are based on trial information.

    Args:
        threshold: The threshold as an integer for this criterion. Ex: If we want to
            generate at most 3 trials, then the threshold is 3.
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: MaxGenerationParallelism
            defaults to setting this to False since we can complete and move on from
            this node without ever reaching its threshold.
        block_gen_if_met: A flag to prevent continued generation from the
            associated GenerationNode if this criterion is met but other criterion
            remain unmet. Ex: ``MinTrials`` has not been met yet, but
            MinTrials has been reached. If this flag is set to true on MinTrials then
            we will raise an error, otherwise we will continue to generate trials
            until ``MinTrials`` is met (thus overriding MinTrials).
        only_in_statuses: A list of trial statuses to filter on when checking the
            criterion threshold.
        not_in_statuses: A list of trial statuses to exclude when checking the
            criterion threshold.
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to when this criterion is met, if it exists.
        use_all_trials_in_exp: A flag to use all trials in the experiment, instead of
            only those generated by the current GenerationNode.
        continue_trial_generation: A flag to indicate that all generation for a given
            trial is not completed, and thus even after transition, the next node will
            continue to generate arms for the same trial. Example usage: in
            ``BatchTrial``s we may  enable generation of arms within a batch from
            different ``GenerationNodes`` by setting this flag to True.
        count_only_trials_with_data: If set to True, only trials with data will be
            counted towards the ``threshold``. Defaults to False.
    """

    def __init__(
        self,
        threshold: int,
        block_transition_if_unmet: bool | None = True,
        block_gen_if_met: bool | None = False,
        only_in_statuses: list[TrialStatus] | None = None,
        not_in_statuses: list[TrialStatus] | None = None,
        transition_to: str | None = None,
        use_all_trials_in_exp: bool | None = False,
        continue_trial_generation: bool | None = False,
        count_only_trials_with_data: bool = False,
    ) -> None:
        self.threshold = threshold
        self.only_in_statuses = only_in_statuses
        self.not_in_statuses = not_in_statuses
        self.use_all_trials_in_exp = use_all_trials_in_exp
        self.count_only_trials_with_data = count_only_trials_with_data
        super().__init__(
            transition_to=transition_to,
            block_transition_if_unmet=block_transition_if_unmet,
            block_gen_if_met=block_gen_if_met,
            continue_trial_generation=continue_trial_generation,
        )

    def experiment_trials_by_status(
        self, experiment: Experiment, statuses: list[TrialStatus]
    ) -> set[int]:
        """Get the trial indices from the entire experiment with the desired
        statuses.

        Args:
            experiment: The experiment associated with this GenerationStrategy.
            statuses: The trial statuses to filter on.
        Returns:
            The trial indices in the experiment with the desired statuses.
        """
        exp_trials_with_statuses = set()
        for status in statuses:
            exp_trials_with_statuses = exp_trials_with_statuses.union(
                experiment.trial_indices_by_status[status]
            )
        return exp_trials_with_statuses

    def all_trials_to_check(self, experiment: Experiment) -> set[int]:
        """All the trials to check from the entire experiment that meet
        all the provided status filters.

        Args:
            experiment: The experiment associated with this GenerationStrategy.
        """
        trials_to_check = set(experiment.trials.keys())
        if self.only_in_statuses is not None:
            trials_to_check = self.experiment_trials_by_status(
                experiment=experiment, statuses=self.only_in_statuses
            )
        # exclude the trials to those not in the specified statuses
        if self.not_in_statuses is not None:
            trials_to_check -= self.experiment_trials_by_status(
                experiment=experiment, statuses=self.not_in_statuses
            )
        return trials_to_check

    def num_contributing_to_threshold(
        self, experiment: Experiment, trials_from_node: set[int]
    ) -> int:
        """Returns the number of trials contributing to the threshold.

        Args:
            experiment: The experiment associated with this GenerationStrategy.
            trials_from_node: The set of trials generated by this GenerationNode.
        """
        all_trials_to_check = self.all_trials_to_check(experiment=experiment)
        if self.count_only_trials_with_data:
            all_trials_to_check = {
                trial_index
                for trial_index in all_trials_to_check
                # TODO[@mgarrard]: determine if we need to actually check data with
                # more granularity, e.g. number of days of data, etc.
                if trial_index in experiment.data_by_trial
            }
        # Some criteria may rely on experiment level data, instead of only trials
        # generated from the node associated with the criterion.
        if self.use_all_trials_in_exp:
            return len(all_trials_to_check)

        return len(trials_from_node.intersection(all_trials_to_check))

    def num_till_threshold(
        self, experiment: Experiment, trials_from_node: set[int]
    ) -> int:
        """Returns the number of trials needed to meet the threshold.

        Args:
            experiment: The experiment associated with this GenerationStrategy.
            trials_from_node: The set of trials generated by this GenerationNode.
        """
        return self.threshold - self.num_contributing_to_threshold(
            experiment=experiment, trials_from_node=trials_from_node
        )

    def is_met(
        self,
        experiment: Experiment,
        curr_node: GenerationNode,
    ) -> bool:
        """Returns if this criterion has been met given its constraints.
        Args:
            experiment: The experiment associated with this GenerationStrategy.
            trials_from_node: The set of trials generated by this GenerationNode.
            block_continued_generation: A flag to prevent continued generation from the
                associated GenerationNode if this criterion is met but other criterion
                remain unmet. Ex: ``MinTrials`` has not been met yet, but
                MinTrials has been reached. If this flag is set to true on MinTrials
                then we will raise an error, otherwise we will continue to generate
                trials until ``MinTrials`` is met (thus overriding MinTrials).
        """
        return (
            self.num_contributing_to_threshold(
                experiment=experiment, trials_from_node=curr_node.trials_from_node
            )
            >= self.threshold
        )


class MaxGenerationParallelism(TrialBasedCriterion):
    """Specific TransitionCriterion implementation which defines the maximum number
    of trials that can simultaneously be in the designated trial statuses. The
    default behavior is to block generation from the associated GenerationNode if the
    threshold is met. This is configured via the `block_gen_if_met` flag being set to
    True. This criterion defaults to not blocking transition to another node via the
    `block_transition_if_unmet` flag being set to False.

    Args:
        threshold: The threshold as an integer for this criterion. Ex: If we want to
            generate at most 3 trials, then the threshold is 3.
        only_in_statuses: A list of trial statuses to filter on when checking the
            criterion threshold.
        not_in_statuses: A list of trial statuses to exclude when checking the
            criterion threshold.
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to when this criterion is met, if it exists.
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: MaxGenerationParallelism
            defaults to setting this to False since we can complete and move on from
            this node without ever reaching its threshold.
        block_gen_if_met: A flag to prevent continued generation from the
            associated GenerationNode if this criterion is met but other criterion
            remain unmet. Ex: ``MinTrials`` has not been met yet, but
            MinTrials has been reached. If this flag is set to true on MinTrials then
            we will raise an error, otherwise we will continue to generate trials
            until ``MinTrials`` is met (thus overriding MinTrials).
        use_all_trials_in_exp: A flag to use all trials in the experiment, instead of
            only those generated by the current GenerationNode.
        continue_trial_generation: A flag to indicate that all generation for a given
            trial is not completed, and thus even after transition, the next node will
            continue to generate arms for the same trial. Example usage: in
            ``BatchTrial``s we may  enable generation of arms within a batch from
            different ``GenerationNodes`` by setting this flag to True. Defaults to
            False for MaxGenerationParallelism since this criterion isn't currently
            used for node -> node or trial -> trial transition.
        count_only_trials_with_data: If set to True, only trials with data will be
            counted towards the ``threshold``. Defaults to False.
    """

    def __init__(
        self,
        threshold: int,
        only_in_statuses: list[TrialStatus] | None = None,
        not_in_statuses: list[TrialStatus] | None = None,
        transition_to: str | None = None,
        block_transition_if_unmet: bool | None = False,
        block_gen_if_met: bool | None = True,
        use_all_trials_in_exp: bool | None = False,
        continue_trial_generation: bool | None = True,
    ) -> None:
        super().__init__(
            threshold=threshold,
            only_in_statuses=only_in_statuses,
            not_in_statuses=not_in_statuses,
            transition_to=transition_to,
            block_gen_if_met=block_gen_if_met,
            block_transition_if_unmet=block_transition_if_unmet,
            use_all_trials_in_exp=use_all_trials_in_exp,
            continue_trial_generation=continue_trial_generation,
        )

    def block_continued_generation_error(
        self,
        node_name: str,
        experiment: Experiment,
        trials_from_node: set[int],
    ) -> None:
        """Raises the appropriate error (should only be called when the
        ``GenerationNode`` is blocked from continued generation). For this
        class, the exception is ``MaxParallelismReachedException``.
        """
        assert self.block_gen_if_met  # Sanity check.
        raise MaxParallelismReachedException(
            node_name=node_name,
            num_running=self.num_contributing_to_threshold(
                experiment=experiment, trials_from_node=trials_from_node
            ),
        )


class MaxTrials(TrialBasedCriterion):
    """
    Simple class to enforce a maximum threshold for the number of trials with the
    designated statuses being generated by a specific GenerationNode. The default
    behavior is to block transition to the next node if the threshold is unmet, but
    not affect continued generation.

    Args:
        threshold: The threshold as an integer for this criterion. Ex: If we want to
            generate at most 3 trials, then the threshold is 3.
        only_in_statuses: A list of trial statuses to filter on when checking the
            criterion threshold.
        not_in_statuses: A list of trial statuses to exclude when checking the
            criterion threshold.
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to when this criterion is met, if it exists.
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: MaxGenerationParallelism
            defaults to setting this to False since we can complete and move on from
            this node without ever reaching its threshold.
        block_gen_if_met: A flag to prevent continued generation from the
            associated GenerationNode if this criterion is met but other criterion
            remain unmet. Ex: MinimumTrialsInStatus has not been met yet, but
            MaxTrials has been reached. If this flag is set to true on MaxTrials then
            we will raise an error, otherwise we will continue to generate trials
            until MinimumTrialsInStatus is met (thus overriding MaxTrials).
        use_all_trials_in_exp: A flag to use all trials in the experiment, instead of
            only those generated by the current GenerationNode.
        continue_trial_generation: A flag to indicate that all generation for a given
            trial is not completed, and thus even after transition, the next node will
            continue to generate arms for the same trial. Example usage: in
            ``BatchTrial``s we may  enable generation of arms within a batch from
            different ``GenerationNodes`` by setting this flag to True.
        count_only_trials_with_data: If set to True, only trials with data will be
            counted towards the ``threshold``. Defaults to False.
    """

    def __init__(
        self,
        threshold: int,
        only_in_statuses: list[TrialStatus] | None = None,
        not_in_statuses: list[TrialStatus] | None = None,
        transition_to: str | None = None,
        block_transition_if_unmet: bool | None = True,
        block_gen_if_met: bool | None = False,
        use_all_trials_in_exp: bool | None = False,
        continue_trial_generation: bool | None = False,
        count_only_trials_with_data: bool = False,
    ) -> None:
        super().__init__(
            threshold=threshold,
            only_in_statuses=only_in_statuses,
            not_in_statuses=not_in_statuses,
            transition_to=transition_to,
            block_gen_if_met=block_gen_if_met,
            block_transition_if_unmet=block_transition_if_unmet,
            use_all_trials_in_exp=use_all_trials_in_exp,
            continue_trial_generation=continue_trial_generation,
            count_only_trials_with_data=count_only_trials_with_data,
        )

    def block_continued_generation_error(
        self,
        node_name: str,
        experiment: Experiment,
        trials_from_node: set[int],
    ) -> None:
        """Raises the appropriate error (should only be called when the
        ``GenerationNode`` is blocked from continued generation). For this
        class, the exception is ``DataRequiredError``.
        """
        assert self.block_gen_if_met  # Sanity check.
        raise DataRequiredError(DATA_REQUIRED_MSG.format(node_name=node_name))


class MinTrials(TrialBasedCriterion):
    """
    Simple class to enforce a minimum threshold for the number of trials with the
    designated statuses being generated by a specific GenerationNode. The default
    behavior is to block transition to the next node if the threshold is unmet, but
    not affect continued generation.

    Args:
        threshold: The threshold as an integer for this criterion. Ex: If we want to
            generate at most 3 trials, then the threshold is 3.
        only_in_statuses: A list of trial statuses to filter on when checking the
            criterion threshold.
        not_in_statuses: A list of trial statuses to exclude when checking the
            criterion threshold.
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to when this criterion is met, if it exists.
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: MaxGenerationParallelism
            defaults to setting this to False since we can complete and move on from
            this node without ever reaching its threshold.
        block_gen_if_met: A flag to prevent continued generation from the
            associated GenerationNode if this criterion is met but other criterion
            remain unmet. Ex: ``MinTrials`` has not been met yet, but
            MinTrials has been reached. If this flag is set to true on MinTrials then
            we will raise an error, otherwise we will continue to generate trials
            until ``MinTrials`` is met (thus overriding MinTrials).
        use_all_trials_in_exp: A flag to use all trials in the experiment, instead of
            only those generated by the current GenerationNode.
        continue_trial_generation: A flag to indicate that all generation for a given
            trial is not completed, and thus even after transition, the next node will
            continue to generate arms for the same trial. Example usage: in
            ``BatchTrial``s we may  enable generation of arms within a batch from
            different ``GenerationNodes`` by setting this flag to True.
        count_only_trials_with_data: If set to True, only trials with data will be
            counted towards the ``threshold``. Defaults to False.
    """

    def __init__(
        self,
        threshold: int,
        only_in_statuses: list[TrialStatus] | None = None,
        not_in_statuses: list[TrialStatus] | None = None,
        transition_to: str | None = None,
        block_transition_if_unmet: bool | None = True,
        block_gen_if_met: bool | None = False,
        use_all_trials_in_exp: bool | None = False,
        continue_trial_generation: bool | None = False,
        count_only_trials_with_data: bool = False,
    ) -> None:
        super().__init__(
            threshold=threshold,
            only_in_statuses=only_in_statuses,
            not_in_statuses=not_in_statuses,
            transition_to=transition_to,
            block_gen_if_met=block_gen_if_met,
            block_transition_if_unmet=block_transition_if_unmet,
            use_all_trials_in_exp=use_all_trials_in_exp,
            continue_trial_generation=continue_trial_generation,
            count_only_trials_with_data=count_only_trials_with_data,
        )

    def block_continued_generation_error(
        self,
        node_name: str,
        experiment: Experiment,
        trials_from_node: set[int],
    ) -> None:
        """Raises the appropriate error (should only be called when the
        ``GenerationNode`` is blocked from continued generation). For this
        class, the exception is ``DataRequiredError``.
        """
        assert self.block_gen_if_met  # Sanity check.
        raise DataRequiredError(DATA_REQUIRED_MSG.format(node_name=node_name))


class MinimumPreferenceOccurances(TransitionCriterion):
    """
    In a preference Experiment (i.e. Metric values may either be zero for No and
    nonzero for Yes) do not transition until a minimum number of both Yes and No
    responses have been received.

    Args:
        metric_name: name of the metric to check for preference occurrences.
        threshold: The threshold as an integer for this criterion. Ex: If we want to
            generate at most 3 trials, then the threshold is 3.
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to when this criterion is met, if it exists.
        block_gen_if_met: A flag to prevent continued generation from the
            associated GenerationNode if this criterion is met but other criterion
            remain unmet. Ex: ``MinTrials`` has not been met yet, but
            MinTrials has been reached. If this flag is set to true on MinTrials then
            we will raise an error, otherwise we will continue to generate trials
            until ``MinTrials`` is met (thus overriding MinTrials).
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: MaxGenerationParallelism
            defaults to setting this to False since we can complete and move on from
            this node without ever reaching its threshold.
    """

    def __init__(
        self,
        metric_name: str,
        threshold: int,
        transition_to: str | None = None,
        block_gen_if_met: bool | None = False,
        block_transition_if_unmet: bool | None = True,
    ) -> None:
        self.metric_name = metric_name
        self.threshold = threshold
        super().__init__(
            transition_to=transition_to,
            block_gen_if_met=block_gen_if_met,
            block_transition_if_unmet=block_transition_if_unmet,
        )

    def is_met(
        self,
        experiment: Experiment,
        curr_node: GenerationNode,
    ) -> bool:
        # TODO: @mgarrard replace fetch_data with lookup_data
        data = experiment.fetch_data(metrics=[experiment.metrics[self.metric_name]])

        count_no = (data.df["mean"] == 0).sum()
        count_yes = (data.df["mean"] != 0).sum()

        return count_no >= self.threshold and count_yes >= self.threshold

    def block_continued_generation_error(
        self,
        node_name: str,
        experiment: Experiment,
        trials_from_node: set[int],
    ) -> None:
        pass


class AuxiliaryExperimentCheck(TransitionCriterion):
    """A class to transition from one GenerationNode to another by checking if certain
    types of Auxiliary Experiment purposes exists.

    A common use case is to use auxiliary_experiment_purposes_to_include to transition
    to a node and auxiliary_experiment_purposes_to_exclude to transition away from it.

    Example usage: In Bayesian optimization with preference exploration (BOPE), we
    check if the preference exploration (PE) auxiliary experiment exists to indicate
    transition to the node that will generate candidates based on the learned
    objective.Since preference exploration is usually conducted after the exploratory
    batch is completed, we do not know at experiment creation time if the PE node
    should be used during the GenerationStrategy.

    Args:
        transition_to: The name of the GenerationNode the GenerationStrategy should
            transition to when this criterion is met, if it exists.
        auxiliary_experiment_purposes_to_include: Optional list of auxiliary experiment
            purposes we expect to have. This can be helpful when need to transition to
            a node based on AuxiliaryExperimentPurpose. Criterion is met when all
            inclusion and exclusion checks pass.
        auxiliary_experiment_purposes_to_exclude: Optional list of auxiliary experiment
            purpose we expect to not have. This can be helpful when need to transition
            out of a node based on AuxiliaryExperimentPurpose. Criterion is met when
            all inclusion and exclusion checks pass.
        block_gen_if_met: A flag to prevent continued generation from the
            associated GenerationNode if this criterion is met but other criterion
            remain unmet. Ex: ``MinTrials`` has not been met yet, but
            MinTrials has been reached. If this flag is set to true on MinTrials then
            we will raise an error, otherwise we will continue to generate trials
            until ``MinTrials`` is met (thus overriding MinTrials).
        block_transition_if_unmet: A flag to prevent the node from completing and
            being able to transition to another node. Ex: MaxGenerationParallelism
            defaults to setting this to False since we can complete and move on from
            this node without ever reaching its threshold.
        continue_trial_generation: A flag to indicate that all generation for a given
            trial is not completed, and thus even after transition, the next node will
            continue to generate arms for the same trial. Example usage: in
            ``BatchTrial``s we may  enable generation of arms within a batch from
            different ``GenerationNodes`` by setting this flag to True.
    """

    def __init__(
        self,
        transition_to: str,
        auxiliary_experiment_purposes_to_include: (
            list[AuxiliaryExperimentPurpose] | None
        ) = None,
        auxiliary_experiment_purposes_to_exclude: (
            list[AuxiliaryExperimentPurpose] | None
        ) = None,
        block_transition_if_unmet: bool | None = True,
        block_gen_if_met: bool | None = False,
        continue_trial_generation: bool | None = False,
    ) -> None:
        super().__init__(
            transition_to=transition_to,
            block_transition_if_unmet=block_transition_if_unmet,
            block_gen_if_met=block_gen_if_met,
            continue_trial_generation=continue_trial_generation,
        )

        if (
            auxiliary_experiment_purposes_to_include is None
            and auxiliary_experiment_purposes_to_exclude is None
        ):
            raise UserInputError(
                f"{self.__class__} cannot have both "
                "`auxiliary_experiment_purposes_to_include` and "
                "`auxiliary_experiment_purposes_to_exclude` be None."
            )
        self.auxiliary_experiment_purposes_to_include = (
            auxiliary_experiment_purposes_to_include
        )
        self.auxiliary_experiment_purposes_to_exclude = (
            auxiliary_experiment_purposes_to_exclude
        )

    def check_aux_exp_purposes(
        self,
        aux_exp_by_purposes: dict[
            AuxiliaryExperimentPurpose, list[AuxiliaryExperiment]
        ],
        include: bool,
        expected_aux_exp_purposes: list[AuxiliaryExperimentPurpose] | None = None,
    ) -> bool:
        """Helper method to check if all elements in expected_aux_exp_purposes
        are in (or not in) aux_exp_purposes"""
        if expected_aux_exp_purposes is not None:
            for purpose in none_throws(expected_aux_exp_purposes):
                purpose_present = (
                    purpose in aux_exp_by_purposes
                    and len(aux_exp_by_purposes[purpose]) > 0
                )
                if purpose_present != include:
                    return False
        return True

    def is_met(
        self,
        experiment: Experiment,
        curr_node: GenerationNode,
    ) -> bool:
        """Check if the experiment has auxiliary experiments for certain purpose."""
        inclusion_check = self.check_aux_exp_purposes(
            aux_exp_by_purposes=experiment.auxiliary_experiments_by_purpose,
            include=True,
            expected_aux_exp_purposes=self.auxiliary_experiment_purposes_to_include,
        )
        exclusion_check = self.check_aux_exp_purposes(
            aux_exp_by_purposes=experiment.auxiliary_experiments_by_purpose,
            include=False,
            expected_aux_exp_purposes=self.auxiliary_experiment_purposes_to_exclude,
        )
        return inclusion_check and exclusion_check

    def block_continued_generation_error(
        self,
        node_name: str,
        experiment: Experiment,
        trials_from_node: set[int],
    ) -> None:
        """Raises the appropriate error (should only be called when the
        ``GenerationNode`` is blocked from continued generation). For this
        class, the exception is ``DataRequiredError``.
        """
        assert self.block_gen_if_met  # Sanity check.
        raise DataRequiredError(
            f"This criterion, {self.criterion_class} has been met but cannot "
            "continue generation from its associated GenerationNode."
        )


# TODO: Deprecate once legacy usecase is updated
class MinimumTrialsInStatus(TransitionCriterion):
    """
    Deprecated and replaced with more flexible MinTrials criterion.
    """

    def __init__(
        self,
        status: TrialStatus,
        threshold: int,
        transition_to: str | None = None,
    ) -> None:
        self.status = status
        self.threshold = threshold
        super().__init__(transition_to=transition_to)

    def is_met(
        self,
        experiment: Experiment,
        curr_node: GenerationNode,
    ) -> bool:
        return len(experiment.trial_indices_by_status[self.status]) >= self.threshold

    def block_continued_generation_error(
        self,
        node_name: str | None,
        experiment: Experiment | None,
        trials_from_node: set[int],
    ) -> None:
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Module for benchmarking Ax algorithms.

Key terms used:

* Replication: 1 run of an optimization loop; (BenchmarkProblem, BenchmarkMethod) pair.
* Test: multiple replications, ran for statistical significance.
* Full run: multiple tests on many (BenchmarkProblem, BenchmarkMethod) pairs.
* Method: (one of) the algorithm(s) being benchmarked.
* Problem: a synthetic function, a surrogate surface, or an ML model, on which
  to assess the performance of algorithms.

"""

import warnings
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import replace
from datetime import datetime
from itertools import product
from logging import Logger, WARNING
from time import time

import numpy as np
import numpy.typing as npt
import pandas as pd
from ax.benchmark.benchmark_method import BenchmarkMethod
from ax.benchmark.benchmark_problem import BenchmarkProblem
from ax.benchmark.benchmark_result import AggregatedBenchmarkResult, BenchmarkResult
from ax.benchmark.benchmark_runner import BenchmarkRunner
from ax.benchmark.benchmark_test_function import BenchmarkTestFunction
from ax.benchmark.methods.sobol import get_sobol_generation_strategy
from ax.core.arm import Arm
from ax.core.experiment import Experiment
from ax.core.map_data import MAP_KEY, MapData
from ax.core.objective import MultiObjective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.search_space import SearchSpace
from ax.core.trial import BaseTrial, Trial
from ax.core.types import TParameterization, TParamValue
from ax.core.utils import get_model_times
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.service.orchestrator import Orchestrator
from ax.service.utils.best_point import get_trace
from ax.service.utils.best_point_mixin import BestPointMixin
from ax.service.utils.orchestrator_options import OrchestratorOptions, TrialType
from ax.utils.common.logger import DEFAULT_LOG_LEVEL, get_logger
from ax.utils.common.random import with_rng_seed
from ax.utils.testing.backend_simulator import BackendSimulator

from pyre_extensions import assert_is_instance, none_throws

logger: Logger = get_logger(__name__)


def update_trials_to_use_sim_time_in_place(
    trials: dict[int, BaseTrial], simulator: BackendSimulator
) -> None:
    """
    Update the start and end times of all trials to be in simulated time
    (represented as datetime objects -- seconds since the start of time).
    """
    fromtimestamp = datetime.fromtimestamp
    for trial_index, trial in trials.items():
        sim_trial = none_throws(
            simulator.get_sim_trial_by_index(trial_index=trial_index)
        )
        trial._time_created = fromtimestamp(
            timestamp=none_throws(sim_trial.sim_queued_time)
        )
        trial._time_staged = fromtimestamp(
            timestamp=none_throws(sim_trial.sim_queued_time)
        )
        trial._time_completed = fromtimestamp(
            timestamp=none_throws(sim_trial.sim_completed_time)
        )
        trial._time_run_started = fromtimestamp(none_throws(sim_trial.sim_start_time))


def compute_score_trace(
    optimization_trace: npt.NDArray, baseline_value: float, optimal_value: float
) -> npt.NDArray:
    """
    Compute a score trace from the optimization trace.

    Score is expressed as a percentage of possible improvement over a baseline.
    A higher score is better.

    Element `i` of the score trace is `optimization_trace[i] - baseline`
    expressed as a percent of `optimal_value - baseline`, where `baseline` is
    `optimization_trace[num_baseline_trials - 1]`. It can be over 100 if values
    better than `optimal_value` are attained or below 0 if values worse than the
    baseline value are attained.

    Args:
        optimization_trace: Objective values. Can be either higher- or
            lower-is-better.
        baseline_value: Value to use as a baseline. Any values that are not
            better than the baseline will receive negative scores.
        optimal_value: The best possible value of the objective; when the
            optimization_trace equals the optimal_value, the score is 100.
    """
    return (
        100 * (optimization_trace - baseline_value) / (optimal_value - baseline_value)
    )


def get_benchmark_runner(
    problem: BenchmarkProblem,
    max_concurrency: int = 1,
    force_use_simulated_backend: bool = False,
) -> BenchmarkRunner:
    """
    Construct a ``BenchmarkRunner`` for the given problem and concurrency.

    If ``max_concurrency > 1`` or if there is a ``sample_runtime_func`` is
    present on ``BenchmarkProblem``, construct a ``SimulatedBenchmarkRunner`` to
    track when trials start and stop.

    Args:
        problem: The ``BenchmarkProblem``; provides a ``BenchmarkTestFunction``
            (used to generate data) and ``step_runtime_function`` (used to
            determine timing for the simulator).
        max_concurrency: The maximum number of trials that can be run concurrently.
            Typically, ``max_pending_trials`` from ``OrchestratorOptions``, which are
            stored on the ``BenchmarkMethod``.
        force_use_simulated_backend: Whether to use a simulated backend even if
            ``max_concurrency`` is 1 and ``problem.step_runtime_function`` is
            None. Recommended for use with a ``BenchmarkMethod`` that uses early
            stopping.
    """

    return BenchmarkRunner(
        test_function=problem.test_function,
        noise_std=problem.noise_std,
        step_runtime_function=problem.step_runtime_function,
        max_concurrency=max_concurrency,
        force_use_simulated_backend=force_use_simulated_backend,
    )


def get_oracle_experiment_from_params(
    problem: BenchmarkProblem,
    dict_of_dict_of_params: Mapping[int, Mapping[str, Mapping[str, TParamValue]]],
) -> Experiment:
    """
    Get a new experiment with the same search space and optimization config
    as those belonging to this problem, but with parameterizations evaluated
    at oracle values (noiseless ground-truth values evaluated at the target task
    and fidelity).

    Args:
        problem: ``BenchmarkProblem`` from which to take a test function for
            generating metrics, as well as a search space and optimization
            config for generating an experiment.
        dict_of_dict_of_params: Keys are trial indices, values are Mappings
            (e.g. dicts) that map arm names to parameterizations.

    Example:
        >>> get_oracle_experiment_from_params(
        ...     problem=problem,
        ...     dict_of_dict_of_params={
        ...         0: {
        ...            "0_0": {"x0": 0.0, "x1": 0.0},
        ...            "0_1": {"x0": 0.3, "x1": 0.4},
        ...         },
        ...         1: {"1_0": {"x0": 0.0, "x1": 0.0}},
        ...     }
        ... )
    """

    experiment = Experiment(
        search_space=problem.search_space,
        optimization_config=problem.optimization_config,
    )

    runner = BenchmarkRunner(test_function=problem.test_function, noise_std=0.0)

    # Silence INFO logs from ax.core.experiment that state "Attached custom
    # parameterizations"
    logger = get_logger("ax.core.experiment")
    original_log_level = logger.level
    logger.setLevel(level="WARNING")

    for trial_index, dict_of_params in dict_of_dict_of_params.items():
        if len(dict_of_params) == 0:
            raise ValueError(
                "Can't create a trial with no arms. Each sublist in "
                "list_of_list_of_params must have at least one element."
            )
        experiment.attach_trial(
            parameterizations=[
                {**parameters, **problem.target_fidelity_and_task}
                for parameters in dict_of_params.values()
            ],
            arm_names=list(dict_of_params.keys()),
        )
        trial = experiment.trials[trial_index]
        metadata = runner.run(trial=trial)
        trial.update_run_metadata(metadata=metadata)
        trial.mark_completed()

    logger.setLevel(level=original_log_level)

    experiment.fetch_data()
    return experiment


def get_benchmark_orchestrator_options(
    method: BenchmarkMethod,
    include_sq: bool = False,
    logging_level: int = DEFAULT_LOG_LEVEL,
) -> OrchestratorOptions:
    """
    Get the ``OrchestratorOptions`` for the given ``BenchmarkMethod``.

    Args:
        method: The ``BenchmarkMethod``.
        include_sq: Whether to include the status quo in each trial.

    Returns:
        ``OrchestratorOptions``
    """
    if method.batch_size is None or method.batch_size > 1 or include_sq:
        trial_type = TrialType.BATCH_TRIAL
    else:
        trial_type = TrialType.TRIAL
    return OrchestratorOptions(
        # No new candidates can be generated while any are pending.
        # If batched, an entire batch must finish before the next can be
        # generated.
        max_pending_trials=method.max_pending_trials,
        # Do not throttle, as is often necessary when polling real endpoints
        init_seconds_between_polls=0,
        min_seconds_before_poll=0,
        trial_type=trial_type,
        batch_size=method.batch_size,
        run_trials_in_batches=method.run_trials_in_batches,
        early_stopping_strategy=method.early_stopping_strategy,
        status_quo_weight=1.0 if include_sq else 0.0,
        logging_level=logging_level,
    )


def _get_oracle_value_of_params(
    params: Mapping[str, TParamValue], problem: BenchmarkProblem
) -> float:
    """
    A roundabout way of getting the value of a parameterization:
    1. Construct an experiment with the parameterization as its only trial,
        using the BenchmarkProblem to get the oracle value of that
        parameterization.
    2. Get the optimization trace of that experiment.
    """
    dummy_experiment = get_oracle_experiment_from_params(
        problem=problem, dict_of_dict_of_params={0: {"0_0": params}}
    )
    (inference_value,) = get_trace(
        experiment=dummy_experiment, optimization_config=problem.optimization_config
    )
    return inference_value


def get_inference_trace(
    trial_completion_order: Sequence[set[int]],
    experiment: Experiment,
    generation_strategy: GenerationStrategy,
    problem: BenchmarkProblem,
) -> npt.NDArray:
    """
    Get the inference trace from a completed experiment.

    The inference trace is the value of the parameterization that would have
    been predicted to be best at each time when a trial completes, using only
    information that would have been available at the time.

    Args:
        trial_completion_order: A list of sets of trial indices, where the first
            element includes the trials that finished first (all at the same
            time), the second element is trials that finished next after that,
            etc.
        experiment: Passed to ``get_trace``.
        generation_strategy: Passed to ``get_trace``.
        problem: Used to get the oracle value of each parameterization.

    """
    completed_trial_idcs: set[int] = set()
    inference_trace = np.full(
        shape=(len(trial_completion_order),), fill_value=float("NaN")
    )

    # Inference trace is not supported for MOO.
    if isinstance(experiment.optimization_config, MultiObjectiveOptimizationConfig):
        return inference_trace

    for i, newly_completed_trials in enumerate(trial_completion_order):
        completed_trial_idcs |= newly_completed_trials
        # Note: Ax's best-point functionality doesn't know to predict at the
        # target task or fidelity, so this won't produce good recommendations in
        # MF/MT settings.
        best_params = get_best_parameters(
            experiment=experiment,
            generation_strategy=generation_strategy,
            trial_indices=completed_trial_idcs,
        )
        if best_params is not None:
            inference_trace[i] = _get_oracle_value_of_params(
                params=best_params, problem=problem
            )
    return inference_trace


def get_best_parameters(
    experiment: Experiment,
    generation_strategy: GenerationStrategy,
    trial_indices: Iterable[int] | None = None,
) -> TParameterization | None:
    """
    Get the most promising point.

    Only SOO is supported. It will return None if no best point can be found.

    Args:
        experiment: The experiment to get the data from. This should contain
            values that would be observed in a realistic setting and not
            contain oracle values.
        generation_strategy: The ``GenerationStrategy`` to use to predict the
            best point.
        trial_indices: Use data from only these trials. If None, use all data.
    """
    result = BestPointMixin._get_best_trial(
        experiment=experiment,
        generation_strategy=generation_strategy,
        trial_indices=trial_indices,
    )
    if result is None:
        # This can happen if no points are predicted to satisfy all outcome
        # constraints.
        return None
    _, params, _ = none_throws(result)
    return params


def get_benchmark_result_from_experiment_and_gs(
    experiment: Experiment,
    generation_strategy: GenerationStrategy,
    problem: BenchmarkProblem,
    seed: int,
    strip_runner_before_saving: bool = True,
) -> BenchmarkResult:
    """
    Parse the ``Experiment`` and ``GenerationStrategy`` into a ``BenchmarkResult``.

    All results are ordered according to ``trial_completion_order``.
    After all trials have been run, the `problem` gets the oracle values of each
    "best" parameter; this yields the ``inference trace``. The cumulative
    maximum of the oracle value of each parameterization tested is the
    ``oracle_trace``.

    Args:
        experiment: The completed ``Experiment`` to extract results from.
        generation_strategy: The ``GenerationStrategy`` used to generate
            ``experiment``; it will be ultimately passed to best-point utilities
            in order to generate the ``inference_trace`` on ``BenchmarkResult``.
        problem: The ``BenchmarkProblem`` used to generate ``experiment``. It
            will be used to extract the oracle values of parameterizations, and
            its ``OptimizationConfig`` is used for identifying best points.
        seed: The seed used to generate ``experiment``.
        strip_runner_before_saving: Whether to write the experiment's runner to
            the returned ``BenchmarkResult``.
    """

    runner = assert_is_instance(experiment.runner, BenchmarkRunner)
    sim_runner = runner.simulated_backend_runner
    if sim_runner is not None:
        trial_indices_by_completion_time: dict[datetime, set[int]] = defaultdict(set)
        for trial_index, trial in experiment.trials.items():
            trial_indices_by_completion_time[none_throws(trial._time_completed)].add(
                trial_index
            )
        trial_completion_order = [
            trial_indices_by_completion_time[k]
            for k in sorted(trial_indices_by_completion_time.keys())
        ]
        cost_trace = np.array(
            [
                completion_time.timestamp()
                for completion_time in sorted(trial_indices_by_completion_time.keys())
            ]
        )
    else:
        trial_completion_order = [{i} for i in range(len(experiment.trials))]
        cost_trace = 1.0 + np.arange(len(experiment.trials), dtype=float)

    # {trial_index: {arm_name: params}}
    dict_of_dict_of_params = {
        new_trial_index: {
            arm.name: arm.parameters
            for old_trial_index in trials
            for arm in experiment.trials[old_trial_index].arms
        }
        for new_trial_index, trials in enumerate(trial_completion_order)
    }

    actual_params_oracle_dummy_experiment = get_oracle_experiment_from_params(
        problem=problem, dict_of_dict_of_params=dict_of_dict_of_params
    )
    oracle_trace = np.array(
        get_trace(
            experiment=actual_params_oracle_dummy_experiment,
            optimization_config=problem.optimization_config,
        )
    )
    inference_trace = get_inference_trace(
        trial_completion_order=trial_completion_order,
        experiment=experiment,
        problem=problem,
        generation_strategy=generation_strategy,
    )

    optimization_trace = (
        inference_trace if problem.report_inference_value_as_trace else oracle_trace
    )

    score_trace = compute_score_trace(
        optimization_trace=optimization_trace,
        optimal_value=problem.optimal_value,
        baseline_value=problem.baseline_value,
    )

    fit_time, gen_time = get_model_times(experiment=experiment)
    if strip_runner_before_saving:
        # Strip runner from experiment before returning, so that the experiment can
        # be serialized (the runner can't be)
        experiment.runner = None

    return BenchmarkResult(
        name=experiment.name,
        seed=seed,
        experiment=experiment,
        oracle_trace=oracle_trace,
        inference_trace=inference_trace,
        optimization_trace=optimization_trace,
        score_trace=score_trace,
        cost_trace=cost_trace,
        fit_time=fit_time,
        gen_time=gen_time,
    )


def run_optimization_with_orchestrator(
    problem: BenchmarkProblem,
    method: BenchmarkMethod,
    seed: int,
    orchestrator_logging_level: int = DEFAULT_LOG_LEVEL,
) -> Experiment:
    """
    Optimize the ``problem`` using the ``method`` and ``Orchestrator``, seeding
    the optimization with ``seed``.

    Args:
        problem: The BenchmarkProblem to test against (can be synthetic or real)
        method: The BenchmarkMethod to test
        seed: The seed to use for this replication.
        orchestrator_logging_level: If >INFO, logs will only appear when unexpected
            things happen. If INFO, logs will update when a trial is completed
            and when an early stopping strategy, if present, decides whether or
            not to continue a trial. If DEBUG, logs additionaly include
            information from a `BackendSimulator`, if present.

    Return:
        ``Experiment`` object.
    """
    sq_arm = (
        None
        if problem.status_quo_params is None
        else Arm(name="status_quo", parameters=problem.status_quo_params)
    )
    orchestrator_options = get_benchmark_orchestrator_options(
        method=method,
        include_sq=sq_arm is not None,
        logging_level=orchestrator_logging_level,
    )
    runner = get_benchmark_runner(
        problem=problem,
        max_concurrency=orchestrator_options.max_pending_trials,
        force_use_simulated_backend=method.early_stopping_strategy is not None,
    )
    experiment = Experiment(
        name=f"{problem.name}|{method.name}_{int(time())}",
        search_space=problem.search_space,
        optimization_config=problem.optimization_config,
        runner=runner,
        status_quo=sq_arm,
        auxiliary_experiments_by_purpose=problem.auxiliary_experiments_by_purpose,
    )

    orchestrator = Orchestrator(
        experiment=experiment,
        generation_strategy=method.generation_strategy.clone_reset(),
        options=orchestrator_options,
    )

    with with_rng_seed(seed=seed), warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="Encountered exception in computing model fit quality",
            category=UserWarning,
            module="ax.adapter.cross_validation",
        )
        orchestrator.run_n_trials(
            max_trials=problem.num_trials, timeout_hours=method.timeout_hours
        )

    sim_runner = runner.simulated_backend_runner
    if sim_runner is not None:
        simulator = sim_runner.simulator
        update_trials_to_use_sim_time_in_place(
            trials=experiment.trials, simulator=simulator
        )

    return experiment


def benchmark_replication(
    problem: BenchmarkProblem,
    method: BenchmarkMethod,
    seed: int,
    strip_runner_before_saving: bool = True,
    orchestrator_logging_level: int = DEFAULT_LOG_LEVEL,
) -> BenchmarkResult:
    """
    Run one benchmarking replication (equivalent to one optimization loop).

    Optimize the ``problem`` using the ``method`` and ``Orchestrator``, seeding
    the optimization with ``seed``. This produces an ``Experiment``. Then parse
    the ``Experiment`` into a ``BenchmarkResult``, extracting traces.


    Args:
        problem: The BenchmarkProblem to test against (can be synthetic or real)
        method: The BenchmarkMethod to test
        seed: The seed to use for this replication.
        strip_runner_before_saving: Whether to strip the runner from the
            experiment before saving it. This enables serialization.
        orchestrator_logging_level: If >INFO, logs will only appear when unexpected
            things happen. If INFO, logs will update when a trial is completed
            and when an early stopping strategy, if present, decides whether or
            not to continue a trial. If DEBUG, logs additionally include
            information from a ``BackendSimulator``, if present.

    Return:
        ``BenchmarkResult`` object.
    """
    experiment = run_optimization_with_orchestrator(
        problem=problem,
        method=method,
        seed=seed,
        orchestrator_logging_level=orchestrator_logging_level,
    )

    benchmark_result = get_benchmark_result_from_experiment_and_gs(
        seed=seed,
        experiment=experiment,
        generation_strategy=method.generation_strategy,
        strip_runner_before_saving=strip_runner_before_saving,
        problem=problem,
    )
    return benchmark_result


def compute_baseline_value_from_sobol(
    optimization_config: OptimizationConfig,
    search_space: SearchSpace,
    test_function: BenchmarkTestFunction,
    target_fidelity_and_task: Mapping[str, TParamValue] | None = None,
    n_repeats: int = 50,
) -> float:
    """
    Compute the `baseline_value` that will be assigned to
    a `BenchmarkProblem`.

    Computed by taking the best of five quasi-random Sobol trials and then
    repeating 50 times. The value is evaluated at the ground truth (noiseless
    and at the target task and fidelity).

    Args:
        optimization_config: Typically, the `optimization_config` of a
            `BenchmarkProblem` (or that will later be used to define a
            `BenchmarkProblem`).
        search_space: Similarly, the `search_space` of a `BenchmarkProblem`.
        test_function: Similarly, the `test_function` of a `BenchmarkProblem`.
        target_fidelity_and_task: Typically, the `target_fidelity_and_task` of a
            `BenchmarkProblem`.
        n_repeats: Number of times to repeat the five Sobol trials.
    """
    gs = get_sobol_generation_strategy()
    method = BenchmarkMethod(generation_strategy=gs)
    target_fidelity_and_task = {} if target_fidelity_and_task is None else {}

    # set up a dummy problem so we can use `benchmark_replication`
    # MOO problems are always higher-is-better because they use hypervolume
    higher_is_better = isinstance(optimization_config.objective, MultiObjective) or (
        not optimization_config.objective.minimize
    )
    dummy_problem = BenchmarkProblem(
        name="dummy",
        optimization_config=optimization_config,
        search_space=search_space,
        num_trials=5,
        test_function=test_function,
        # Optimal value and baseline value are only used
        # to compute the score_trace, which we don't use here.
        # The order of baseline and optimal value needs to be correct,
        # though, as a ValueError is raised otherwise.
        optimal_value=1.0 if higher_is_better else -1.0,
        baseline_value=0.0,
        target_fidelity_and_task=target_fidelity_and_task,
    )

    values = np.full(n_repeats, np.nan)
    for i in range(n_repeats):
        result = benchmark_replication(
            problem=dummy_problem,
            method=method,
            seed=i,
            orchestrator_logging_level=WARNING,
        )
        values[i] = result.optimization_trace[-1]

    return values.mean()


def benchmark_one_method_problem(
    problem: BenchmarkProblem,
    method: BenchmarkMethod,
    seeds: Iterable[int],
    orchestrator_logging_level: int = DEFAULT_LOG_LEVEL,
) -> AggregatedBenchmarkResult:
    return AggregatedBenchmarkResult.from_benchmark_results(
        results=[
            benchmark_replication(
                problem=problem,
                method=method,
                seed=seed,
                orchestrator_logging_level=orchestrator_logging_level,
            )
            for seed in seeds
        ]
    )


def benchmark_multiple_problems_methods(
    problems: Iterable[BenchmarkProblem],
    methods: Iterable[BenchmarkMethod],
    seeds: Iterable[int],
    orchestrator_logging_level: int = DEFAULT_LOG_LEVEL,
) -> list[AggregatedBenchmarkResult]:
    """
    For each `problem` and `method` in the Cartesian product of `problems` and
    `methods`, run the replication on each seed in `seeds` and get the results
    as an `AggregatedBenchmarkResult`, then return a list of each
    `AggregatedBenchmarkResult`.
    """
    return [
        benchmark_one_method_problem(
            problem=p,
            method=m,
            seeds=seeds,
            orchestrator_logging_level=orchestrator_logging_level,
        )
        for p, m in product(problems, methods)
    ]


def get_opt_trace_by_steps(experiment: Experiment) -> npt.NDArray:
    """
    Transform an optimization trace in the standard format produced by
    `benchmark_replication`, with one element per trial completion, into a trace
    that is in terms of steps, with one element added each time a step
    completes.

    Args:
        experiment: An experiment produced by `benchmark_replication`; it must
            have `BenchmarkTrialMetadata` (as produced by `BenchmarkRunner`) for
            each trial, and its data must be `MapData`.
    """
    optimization_config = none_throws(experiment.optimization_config)

    if optimization_config.is_moo_problem:
        raise NotImplementedError(
            "Cumulative epochs only supported for single objective problems."
        )
    if len(optimization_config.outcome_constraints) > 0:
        raise NotImplementedError(
            "Cumulative epochs not supported for problems with outcome constraints."
        )

    objective_name = optimization_config.objective.metric.name
    data = assert_is_instance(experiment.lookup_data(), MapData)
    map_df = data.map_df

    # Has timestamps; needs to be merged with map_df because it contains
    # data on epochs that didn't actually run due to early stopping, and we need
    # to know which actually ran
    def _get_df(trial: Trial) -> pd.DataFrame:
        """
        Get the (virtual) time each epoch finished at.
        """
        metadata = trial.run_metadata["benchmark_metadata"]
        backend_simulator = none_throws(metadata.backend_simulator)
        # Data for the first metric, which is the only metric
        df = next(iter(metadata.dfs.values()))
        start_time = backend_simulator.get_sim_trial_by_index(
            trial.index
        ).sim_start_time
        df["time"] = df["virtual runtime"] + start_time
        return df

    with_timestamps = pd.concat(
        (
            _get_df(trial=assert_is_instance(trial, Trial))
            for trial in experiment.trials.values()
        ),
        axis=0,
        ignore_index=True,
    )[["trial_index", MAP_KEY, "time"]]

    df = (
        map_df.loc[
            map_df["metric_name"] == objective_name,
            ["trial_index", "arm_name", "mean", MAP_KEY],
        ]
        .merge(with_timestamps, how="left")
        .sort_values("time", ignore_index=True)
    )
    return (
        df["mean"].cummin()
        if optimization_config.objective.minimize
        else df["mean"].cummax()
    ).to_numpy()


def get_benchmark_result_with_cumulative_steps(
    result: BenchmarkResult,
    optimal_value: float,
    baseline_value: float,
) -> BenchmarkResult:
    """
    Replaces the cost trace with the cumulative number of steps run and
    recomputes the optimization trace accordingly, using
    `get_opt_trace_by_steps`.
    """

    experiment = none_throws(result.experiment)
    opt_trace = get_opt_trace_by_steps(experiment=experiment)
    return replace(
        result,
        optimization_trace=opt_trace,
        cost_trace=np.arange(1, len(opt_trace) + 1, dtype=int),
        # Empty
        oracle_trace=np.full(len(opt_trace), np.nan),
        inference_trace=np.full(len(opt_trace), np.nan),
        score_trace=compute_score_trace(
            optimization_trace=opt_trace,
            baseline_value=baseline_value,
            optimal_value=optimal_value,
        ),
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections import defaultdict
from typing import Optional, TYPE_CHECKING

import numpy as np
from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.standardize_y import compute_standardization_parameters
from ax.core.observation import Observation, ObservationFeatures, separate_observations
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.parameter import ChoiceParameter
from ax.core.search_space import SearchSpace
from ax.core.types import TParamValue
from ax.generators.types import TConfig
from pyre_extensions import assert_is_instance, none_throws


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class StratifiedStandardizeY(Transform):
    """Standardize Y, separately for each metric and for each value of a
    ChoiceParameter.

    The name of the parameter by which to stratify the standardization can be
    specified in config["parameter_name"]. If not specified, will use a task
    parameter if search space contains exactly 1 task parameter, and will raise
    an exception otherwise.

    The stratification parameter must be fixed during generation if there are
    outcome constraints, in order to apply the standardization to the
    constraints.

    Transform is done in-place.
    """

    requires_data_for_initialization: bool = True

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        """Initialize StratifiedStandardizeY.

        Args:
            search_space: The search space of the experiment.
            observations: A list of observations from the experiment.
            experiment_data: A container for the parameterizations, metadata and
                observations for the trials in the experiment.
                Constructed using ``extract_experiment_data``.
            adapter: Adapter for referencing experiment, status quo, etc.
            config: A dictionary of options that may contain a "parameter_name" key
                specifying the name of the parameter to use for stratification and a
                "strata_mapping" key that corresponds to a dictionary that maps
                parameter values to strata for standardization. The strata can be
                of type bool, int, str, or float.

        """
        assert search_space is not None, "StratifiedStandardizeY requires search space"
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        # Get parameter name for standardization.
        self.strata_mapping = None  # pyre-ignore [8]
        if config is not None and "parameter_name" in config:
            # pyre: Attribute `p_name` declared in class `ax.adapter.
            # pyre: transforms.stratified_standardize_y.
            # pyre: StratifiedStandardizeY` has type `str` but is used as type
            # pyre-fixme[8]: `typing.Union[float, int, str]`.
            self.p_name: str = config["parameter_name"]
            strat_p = search_space.parameters[self.p_name]
            if not isinstance(strat_p, ChoiceParameter):
                raise ValueError(f"{self.p_name} not a ChoiceParameter")
            if "strata_mapping" in config:
                # pyre-ignore [8]
                self.strata_mapping: dict[
                    bool | float | int | str, bool | float | int | str
                ] = config["strata_mapping"]
                if set(strat_p.values) != set(self.strata_mapping.keys()):
                    raise ValueError(
                        f"{self.p_name} values {strat_p.values} do not match "
                        f"strata_mapping keys {self.strata_mapping.keys()}."
                    )
        else:
            # See if there is a task parameter
            task_parameters = [
                p.name
                for p in search_space.parameters.values()
                if isinstance(p, ChoiceParameter) and p.is_task
            ]
            if len(task_parameters) == 0:
                raise ValueError(
                    "Must specify parameter for stratified standardization. This can "
                    "happen if TrialAsTask is a no-op, due to there only being a single"
                    " task level."
                )
            elif len(task_parameters) != 1:
                raise ValueError(
                    "Must specify which task parameter to use for stratified "
                    "standardization"
                )
            self.p_name = task_parameters[0]
        if self.strata_mapping is None:
            strat_p = assert_is_instance(
                search_space.parameters[self.p_name], ChoiceParameter
            )
            # pyre-ignore [8]
            self.strata_mapping = {v: v for v in strat_p.values}
        # Compute means and SDs
        if observations is not None:
            observation_features, observation_data = separate_observations(observations)
            Ys: defaultdict[tuple[str, TParamValue], list[float]] = defaultdict(list)
            for j, obsd in enumerate(observation_data):
                v = none_throws(observation_features[j].parameters[self.p_name])
                strata = self.strata_mapping[v]
                for i, m in enumerate(obsd.metric_names):
                    Ys[(m, strata)].append(obsd.means[i])
        else:
            experiment_data = none_throws(experiment_data)
            if len(experiment_data.observation_data.index.names) > 2:
                raise NotImplementedError(
                    "StratifiedStandardizeY does not support experiment data with "
                    "map keys."
                )
            strata = (
                experiment_data.arm_data[self.p_name]
                .map(self.strata_mapping)
                .rename("strata")
            )
            means = experiment_data.observation_data["mean"]
            Ys: dict[tuple[str, TParamValue], list[float]] = {}
            # Group means by strata values and extract corresponding Ys.
            for strata_value, group in means.groupby(strata.loc[means.index]):
                for m in means.columns:
                    Ys[(m, strata_value)] = group[m].dropna().values.tolist()

        # Expected `DefaultDict[typing.Union[str, typing.Tuple[str,
        # Optional[typing.Union[bool, float, str]]]], List[float]]` for 1st anonymous
        # parameter to call
        # `ax.adapter.transforms.standardize_y.compute_standardization_parameters`
        # but got `DefaultDict[typing.Tuple[str, Optional[typing.Union[bool, float,
        # str]]], List[float]]`.
        # pyre-fixme[6]: Expected `DefaultDict[Union[str, Tuple[str, Optional[Union[b...
        # pyre-fixme[4]: Attribute must be annotated.
        self.Ymean, self.Ystd = compute_standardization_parameters(Ys)

    def transform_observations(
        self,
        observations: list[Observation],
    ) -> list[Observation]:
        # Transform observations
        for obs in observations:
            v = none_throws(obs.features.parameters[self.p_name])
            strata = self.strata_mapping[v]
            means = np.array([self.Ymean[(m, strata)] for m in obs.data.metric_names])
            stds = np.array([self.Ystd[(m, strata)] for m in obs.data.metric_names])
            obs.data.means = (obs.data.means - means) / stds
            obs.data.covariance /= np.dot(stds[:, None], stds[:, None].transpose())
        return observations

    def transform_optimization_config(
        self,
        optimization_config: OptimizationConfig,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        fixed_features: ObservationFeatures | None = None,
    ) -> OptimizationConfig:
        if len(optimization_config.all_constraints) == 0:
            return optimization_config
        if fixed_features is None or self.p_name not in fixed_features.parameters:
            raise ValueError(
                f"StratifiedStandardizeY transform requires {self.p_name} to be fixed "
                "during generation."
            )
        v = none_throws(fixed_features.parameters[self.p_name])
        strata = self.strata_mapping[v]
        for c in optimization_config.all_constraints:
            if c.relative:
                raise ValueError(
                    "StratifiedStandardizeY transform does not support relative "
                    f"constraint {c}"
                )
            c.bound = (c.bound - self.Ymean[(c.metric.name, strata)]) / self.Ystd[
                (c.metric.name, strata)
            ]
        return optimization_config

    def untransform_observations(
        self,
        observations: list[Observation],
    ) -> list[Observation]:
        for obs in observations:
            v = none_throws(obs.features.parameters[self.p_name])
            strata = self.strata_mapping[v]
            means = np.array([self.Ymean[(m, strata)] for m in obs.data.metric_names])
            stds = np.array([self.Ystd[(m, strata)] for m in obs.data.metric_names])
            obs.data.means = obs.data.means * stds + means
            obs.data.covariance *= np.dot(stds[:, None], stds[:, None].transpose())
        return observations

    def untransform_outcome_constraints(
        self,
        outcome_constraints: list[OutcomeConstraint],
        fixed_features: ObservationFeatures | None = None,
    ) -> list[OutcomeConstraint]:
        if fixed_features is None or self.p_name not in fixed_features.parameters:
            raise ValueError(
                f"StratifiedStandardizeY requires {self.p_name} to be fixed here"
            )
        v = none_throws(fixed_features.parameters[self.p_name])
        strata = self.strata_mapping[v]
        for c in outcome_constraints:
            if c.relative:
                raise ValueError(
                    "StratifiedStandardizeY does not support relative constraints"
                )
            c.bound = float(
                c.bound * self.Ystd[(c.metric.name, strata)]
                + self.Ymean[(c.metric.name, strata)]
            )
        return outcome_constraints

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        strata = (
            experiment_data.arm_data[self.p_name]
            .map(self.strata_mapping)
            .rename("strata")
        )
        obs_data = experiment_data.observation_data
        mean_data = obs_data["mean"]
        # Get strata values for each row by matching (trial_index, arm_name).
        strata = strata.loc[[v[:2] for v in mean_data.index.values]]
        for metric in mean_data:
            means = strata.apply(lambda x: self.Ymean[(metric, x)])  # noqa B023
            stds = strata.apply(lambda x: self.Ystd[(metric, x)])  # noqa B023
            obs_data[("mean", metric)] = (obs_data[("mean", metric)] - means) / stds
            if obs_data[("sem", metric)].isnull().all():
                # If SEM is NaN, we don't need to transform it.
                continue
            obs_data[("sem", metric)] = obs_data[("sem", metric)] / stds
        return ExperimentData(
            arm_data=experiment_data.arm_data, observation_data=obs_data
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Callable, Iterable, Mapping
from inspect import Parameter, signature
from logging import Logger
from typing import Any

from ax.utils.common.logger import get_logger

logger: Logger = get_logger(__name__)

TKwargs = dict[str, Any]


def consolidate_kwargs(
    kwargs_iterable: Iterable[Mapping[str, Any] | None], keywords: Iterable[str]
) -> dict[str, Any]:
    """Combine an iterable of kwargs into a single dict of kwargs, where kwargs
    by duplicate keys that appear later in the iterable get priority over the
    ones that appear earlier and only kwargs referenced in keywords will be
    used. This allows to combine somewhat redundant sets of kwargs, where a
    user-set kwarg, for instance, needs to override a default kwarg.

    >>> consolidate_kwargs(
    ...     kwargs_iterable=[{'a': 1, 'b': 2}, {'b': 3, 'c': 4, 'd': 5}],
    ...     keywords=['a', 'b', 'd']
    ... )
    {'a': 1, 'b': 3, 'd': 5}
    """
    all_kwargs = {}
    for kwargs in kwargs_iterable:
        if kwargs is not None:
            all_kwargs.update({kw: p for kw, p in kwargs.items() if kw in keywords})
    return all_kwargs


def get_function_argument_names(
    function: Callable,
    omit: list[str] | None = None,
) -> list[str]:
    """Extract parameter names from function signature."""
    omit = omit or []
    return [p for p in signature(function).parameters.keys() if p not in omit]


def get_function_default_arguments(function: Callable) -> dict[str, Any]:
    """Extract default arguments from function signature."""
    params = signature(function).parameters
    return {
        kw: p.default for kw, p in params.items() if p.default is not Parameter.empty
    }


def warn_on_kwargs(callable_with_kwargs: Callable, **kwargs: Any) -> None:
    """Log a warning when a decoder function receives unexpected kwargs.

    NOTE: This mainly caters to the use case where an older version of Ax is
    used to decode objects, serialized to JSON by a newer version of Ax (and
    therefore potentially containing new fields). In that case, the decoding
    function should not fail when encountering those additional fields, but
    rather just ignore them and log a warning using this function.
    """
    if kwargs:
        logger.warning(
            "Found unexpected kwargs: %s while calling %s "
            "from JSON. These kwargs will be ignored.",
            kwargs,
            callable_with_kwargs,
        )


def filter_kwargs(function: Callable, **kwargs: Any) -> Any:
    """Filter out kwargs that are not applicable for a given function.
    Return a copy of given kwargs dict with only the required kwargs."""
    return {k: v for k, v in kwargs.items() if k in signature(function).parameters}



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections import defaultdict
from copy import deepcopy

import numpy as np
from ax.adapter.base import Adapter
from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.core.observation import Observation, ObservationData
from ax.core.search_space import SearchSpace
from ax.generators.types import TConfig
from pyre_extensions import assert_is_instance, none_throws


class MergeRepeatedMeasurements(Transform):
    """Merge repeated measurements for to obtain one observation per arm.

    Repeated measurements are merged via inverse variance weighting (e.g. over
    different trials). This intentionally ignores the trial index and assumes
    stationarity.

    TODO: Support inverse variance weighting correlated outcomes (full covariance).

    Note: this is not reversible.
    """

    requires_data_for_initialization: bool = True

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Adapter | None = None,
        config: TConfig | None = None,
    ) -> None:
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        # create a mapping of arm_key -> {metric_name: {means: [], vars: []}}
        arm_to_multi_obs: defaultdict[
            str, defaultdict[str, defaultdict[str, list[float]]]
        ] = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        if experiment_data is not None:
            metrics = experiment_data.metric_names
            for arm_name, df in experiment_data.observation_data.groupby(
                level="arm_name"
            ):
                for m in metrics:
                    # Get the subset of the df for the metric.
                    df_m = df[[("mean", m), ("sem", m)]].dropna(
                        axis=0, subset=[("mean", m)]
                    )
                    if any(df_m[("sem", m)].isna()):
                        raise NotImplementedError(
                            "All metrics must have noise observations."
                        )
                    arm_to_multi_obs[arm_name][m]["means"].extend(
                        df_m[("mean", m)].tolist()
                    )
                    arm_to_multi_obs[arm_name][m]["vars"].extend(
                        (df_m[("sem", m)] ** 2).tolist()
                    )
        else:
            for obs in none_throws(observations):
                if (arm_name := obs.arm_name) is None:
                    # Since the transform will be initialized with Adapter training
                    # data, all observations will have arm names.
                    raise NotImplementedError("All observations must have arm names.")
                # TODO: support inverse variance weighting for multivariate
                # distributions (full covariance).
                obsd = obs.data
                diag = np.diag(np.diag(obsd.covariance))
                if np.any(np.isnan(obsd.covariance)):
                    raise NotImplementedError(
                        "All metrics must have noise observations."
                    )
                elif ~np.all(obsd.covariance == diag):
                    raise NotImplementedError(
                        "Only independent metrics are currently supported."
                    )
                for i, m in enumerate(obsd.metric_names):
                    arm_to_multi_obs[arm_name][m]["means"].append(obsd.means[i])
                    arm_to_multi_obs[arm_name][m]["vars"].append(obsd.covariance[i, i])

        self.arm_to_merged: defaultdict[str, dict[str, dict[str, float]]] = defaultdict(
            dict
        )
        for k, metric_dict in arm_to_multi_obs.items():
            for m, v in metric_dict.items():
                # inverse variance weighting
                var = np.array(v["vars"])
                means = np.array(v["means"])
                noiseless = var == 0
                if np.any(noiseless):
                    noiseless_means = means[noiseless]
                    if (noiseless_means.shape[0] > 1) and (
                        not np.all(noiseless_means[1:] == noiseless_means[0])
                    ):
                        raise ValueError(
                            "All repeated arms with noiseless measurements "
                            "must have the same means."
                        )
                    self.arm_to_merged[k][m] = {
                        "mean": noiseless_means[0],
                        "var": 0.0,
                    }
                else:
                    inv_var = 1 / np.array(var)
                    inv_sum_inv_var = 1 / np.sum(inv_var)
                    weights = inv_var * inv_sum_inv_var
                    self.arm_to_merged[k][m] = {
                        "mean": np.sum(means * weights),
                        "var": inv_sum_inv_var,
                    }

    def transform_observations(
        self,
        observations: list[Observation],
    ) -> list[Observation]:
        new_observations = []
        arm_to_merged = deepcopy(self.arm_to_merged)
        for obs in observations:
            arm_name = obs.arm_name
            if arm_name not in arm_to_merged:
                continue
            arm_name = none_throws(arm_name)
            metric_dict = arm_to_merged.pop(arm_name)
            obsd = obs.data
            merged_means = np.zeros(len(obsd.metric_names))
            merged_covariance = np.zeros(
                (len(obsd.metric_names), len(obsd.metric_names))
            )
            for i, m in enumerate(obsd.metric_names):
                merged_metric = metric_dict[m]
                merged_means[i] = merged_metric["mean"]
                merged_covariance[i, i] = merged_metric["var"]
            new_obsd = ObservationData(
                metric_names=obsd.metric_names,
                means=merged_means,
                covariance=merged_covariance,
            )
            new_obs = Observation(
                features=obs.features,
                data=new_obsd,
                arm_name=arm_name,
            )
            new_observations.append(new_obs)
        return new_observations

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        # Transformed arm data will retain the first occurance for each arm
        # in self.arm_to_merged.
        arm_data = experiment_data.arm_data
        arm_data = arm_data.loc[
            arm_data.index.get_level_values("arm_name").isin(self.arm_to_merged.keys())
        ]
        arm_data = arm_data[
            ~arm_data.index.get_level_values("arm_name").duplicated(keep="first")
        ]
        # Observation data should also retain only the first occurance, but the actual
        # data will be overwritten.
        observation_data = experiment_data.observation_data
        observation_data = observation_data.loc[
            observation_data.index.get_level_values("arm_name").isin(
                self.arm_to_merged.keys()
            )
        ]
        observation_data = observation_data[
            ~observation_data.index.get_level_values("arm_name").duplicated(
                keep="first"
            )
        ]
        for index in observation_data.index:
            arm_name = assert_is_instance(index, tuple)[1]
            metric_dict = self.arm_to_merged[arm_name]
            for m in metric_dict:
                observation_data.loc[index, ("mean", m)] = metric_dict[m]["mean"]
                observation_data.loc[index, ("sem", m)] = metric_dict[m]["var"] ** 0.5
        return ExperimentData(
            arm_data=arm_data,
            observation_data=observation_data,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Iterator
from copy import deepcopy
from unittest.mock import ANY

from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.metadata_to_float import MetadataToFloat
from ax.core.observation import ObservationFeatures, observations_from_data
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.exceptions.core import DataRequiredError
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_experiment_with_observations
from pandas.testing import assert_frame_equal
from pyre_extensions import assert_is_instance


WIDTHS = [2.0, 4.0, 8.0]
HEIGHTS = [4.0, 2.0, 8.0]
STEPS_ENDS = [1, 5, 3]


def _enumerate() -> Iterator[tuple[float, float, float]]:
    yield from (
        (width, height, float(i + 1))
        for (width, height, steps_end) in zip(WIDTHS, HEIGHTS, STEPS_ENDS)
        for i in range(steps_end)
    )


class MetadataToFloatTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    name="width",
                    parameter_type=ParameterType.FLOAT,
                    lower=1,
                    upper=20,
                ),
                RangeParameter(
                    name="height",
                    parameter_type=ParameterType.FLOAT,
                    lower=1,
                    upper=20,
                ),
            ]
        )
        self.experiment = get_experiment_with_observations(
            observations=[[0.0] for _ in range(sum(STEPS_ENDS))],
            search_space=self.search_space,
            parameterizations=[
                {"width": w, "height": h}
                for steps_end, w, h in zip(STEPS_ENDS, WIDTHS, HEIGHTS)
                for _ in range(steps_end)
            ],
            candidate_metadata=[
                {
                    "foo": 42,
                    "bar": 3.0 * steps,
                }
                for steps_end, _w, _h in zip(STEPS_ENDS, WIDTHS, HEIGHTS)
                for steps in range(1, steps_end + 1)
            ],
        )
        self.observations = observations_from_data(
            experiment=self.experiment, data=self.experiment.lookup_data()
        )
        self.experiment_data = extract_experiment_data(
            experiment=self.experiment,
            data_loader_config=DataLoaderConfig(),
        )

        self.t = MetadataToFloat(
            observations=self.observations,
            config={
                "parameters": {"bar": {"digits": 4}},
            },
        )
        self.t2 = MetadataToFloat(
            experiment_data=self.experiment_data,
            config={
                "parameters": {"bar": {"digits": 4}},
            },
        )

    def test_Init(self) -> None:
        for t in (self.t, self.t2):
            self.assertEqual(len(t._parameter_list), 1)
            p = t._parameter_list[0]
            # check that the parameter options are specified in a sensible manner
            # by default if the user does not specify them explicitly
            self.assertEqual(p.name, "bar")
            self.assertEqual(p.parameter_type, ParameterType.FLOAT)
            self.assertEqual(p.lower, 3.0)
            self.assertEqual(p.upper, 15.0)
            self.assertFalse(p.log_scale)
            self.assertFalse(p.logit_scale)
            self.assertEqual(p.digits, 4)
            self.assertFalse(p.is_fidelity)
            self.assertIsNone(p.target_value)

        with self.assertRaisesRegex(DataRequiredError, "requires non-empty data"):
            MetadataToFloat(search_space=None, observations=None)
        with self.assertRaisesRegex(DataRequiredError, "requires non-empty data"):
            MetadataToFloat(search_space=None)

    def test_TransformSearchSpace(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)

        self.assertSetEqual(
            set(ss2.parameters.keys()),
            {"height", "width", "bar"},
        )

        p = assert_is_instance(ss2.parameters["bar"], RangeParameter)

        self.assertEqual(p.name, "bar")
        self.assertEqual(p.parameter_type, ParameterType.FLOAT)
        self.assertEqual(p.lower, 3.0)
        self.assertEqual(p.upper, 15.0)
        self.assertFalse(p.log_scale)
        self.assertFalse(p.logit_scale)
        self.assertEqual(p.digits, 4)
        self.assertFalse(p.is_fidelity)
        self.assertIsNone(p.target_value)

    def test_TransformObservationFeatures(self) -> None:
        observation_features = [obs.features for obs in self.observations]
        obs_ft2 = deepcopy(observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)

        self.assertEqual(
            obs_ft2,
            [
                ObservationFeatures(
                    trial_index=trial_index,
                    parameters={
                        "width": width,
                        "height": height,
                        "bar": 3.0 * steps,
                    },
                    metadata={"foo": 42, Keys.TRIAL_COMPLETION_TIMESTAMP: ANY},
                )
                for trial_index, (width, height, steps) in enumerate(_enumerate())
            ],
        )
        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, observation_features)

    def test_transform_experiment_data(self) -> None:
        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(self.experiment_data)
        )
        # Check that arm data now has a new column for the transformed parameter.
        expected_bar_values = [
            3.0 * s for steps in STEPS_ENDS for s in range(1, steps + 1)
        ]
        self.assertEqual(transformed_data.arm_data["bar"].tolist(), expected_bar_values)
        # Remaining columns are unchanged.
        assert_frame_equal(
            transformed_data.arm_data.drop(columns=["bar"]),
            self.experiment_data.arm_data,
        )
        # Observation data is not changed.
        assert_frame_equal(
            transformed_data.observation_data,
            self.experiment_data.observation_data,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import io
import json
import os
import re
import shutil
import subprocess
import uuid
from pathlib import Path
from typing import Union

import mdformat
import nbformat
import pandas as pd
from nbformat.notebooknode import NotebookNode

SCRIPTS_DIR = Path(__file__).parent.resolve()
LIB_DIR = SCRIPTS_DIR.parent.resolve()

WEBSITE_DIR = LIB_DIR.joinpath("website")
DOCS_DIR = LIB_DIR.joinpath("docs")
TUTORIALS_DIR = DOCS_DIR.joinpath("tutorials")

ORGANIZATION = "facebook"
PROJECT = "ax"

# Data display priority. Below lists the priority for displaying data from cell outputs.
# Cells can output many different items, and some will output a fallback display, e.g.
# text/plain if text/html is not working. The below priorities help ensure the output in
# the MDX file shows the best representation of the cell output.
priorities = [
    "error",
    "text/markdown",
    "image/png",  # matplotlib output.
    "application/vnd.jupyter.widget-view+json",  # tqdm progress bars.
    "application/vnd.plotly.v1+json",  # Plotly
    "text/html",
    "stream",
    "text/plain",
]


def load_nb_metadata() -> dict[str, dict[str, str]]:
    """
    Load the metadata and list of notebooks that are to be converted to MDX.

    Args:
        None

    Returns:
        Dict[str, Dict[str, str]]: A dictionary of metadata needed to convert notebooks
            to MDX. Only those notebooks that are listed in the `tutorials.json` file
            will be included in the Docusaurus MDX output.
    """
    tutorials_json_path = WEBSITE_DIR.joinpath("tutorials.json")
    with tutorials_json_path.open("r") as f:
        tutorials_data = json.load(f)
    # flatten tutorial config, we handle the nested structure in sidebars.js
    tutorial_configs = [
        config for category in tutorials_data.values() for config in category
    ]
    return tutorial_configs


def load_notebook(path: Path) -> NotebookNode:
    """
    Load the given notebook into memory.

    Args:
        path (Path): Path to the Jupyter notebook.

    Returns:
        NotebookNode: `nbformat` object, which contains all the notebook cells in it.
    """
    with path.open("r") as f:
        nb_str = f.read()
        nb = nbformat.reads(nb_str, nbformat.NO_CONVERT)
    return nb


def create_folders(path: Path) -> tuple[str, Path]:
    """
    Create asset folders for the tutorial.

    Args:
        path (Path): Path to the Jupyter notebook.

    Returns:
        Tuple[str, Path]: Returns a tuple with the filename to use for the MDX file
            and the path for the MDX assets folder.
    """
    tutorial_folder_name = path.stem
    filename = "".join([token.title() for token in tutorial_folder_name.split("_")])
    tutorial_folder = TUTORIALS_DIR.joinpath(tutorial_folder_name)
    assets_folder = tutorial_folder / "assets"
    img_folder = assets_folder / "img"
    plot_data_folder = assets_folder / "plot_data"
    if not tutorial_folder.exists():
        tutorial_folder.mkdir(parents=True, exist_ok=True)
    if not img_folder.exists():
        img_folder.mkdir(parents=True, exist_ok=True)
    if not plot_data_folder.exists():
        plot_data_folder.mkdir(parents=True, exist_ok=True)
    return filename, assets_folder


def create_frontmatter(path: Path, nb_metadata: dict[str, dict[str, str]]) -> str:
    """
    Create frontmatter for the resulting MDX file.

    The frontmatter is the data between the `---` lines in an MDX file.

    Args:
        path (Path): Path to the Jupyter notebook.
        nb_metadata (Dict[str, Dict[str, str]]): The metadata associated with the given
            notebook. Metadata is defined in the `tutorials.json` file.

    Returns:
        str: MDX formatted frontmatter.
    """
    # Add the frontmatter to the MDX string. This is the part between the `---` lines
    # that define the tutorial sidebar_label information.
    frontmatter_delimiter = ["---"]
    frontmatter = [
        f'"{key}": "{value}"'
        for key, value in {
            "title": metadata["title"],
            "sidebar_label": metadata["title"],
        }.items()
    ]
    frontmatter = "\n".join(frontmatter_delimiter + frontmatter + frontmatter_delimiter)
    mdx = mdformat.text(frontmatter, options={"wrap": 88}, extensions={"myst"})
    return f"{mdx}\n"


def create_imports() -> str:
    """
    Create the imports needed for displaying buttons, and interactive plots in MDX.

    Returns:
        str: MDX formatted imports.
    """
    link_btn = "@site/src/components/LinkButtons.jsx"
    cell_out = "@site/src/components/CellOutput.jsx"
    plot_out = "@site/src/components/Plotting.jsx"
    imports = f'import LinkButtons from "{link_btn}";\n'
    imports += f'import CellOutput from "{cell_out}";\n'
    imports += f'import {{PlotlyFigure}} from "{plot_out}";\n'
    return f"{imports}\n"


def get_current_git_tag() -> str | None:
    """
    Retrieve the current Git tag if the current commit is tagged.

    Returns:
        Optional[str]: The current Git tag as a string if available, otherwise None.
    """
    try:
        tag = (
            subprocess.check_output(
                ["git", "describe", "--tags", "--exact-match"], stderr=subprocess.STDOUT
            )
            .strip()
            .decode("utf-8")
        )
        return tag
    except subprocess.CalledProcessError:
        return None


def create_buttons(
    nb_metadata: dict[str, dict[str, str]],
) -> str:
    """
    Create buttons that link to Colab and GitHub for the tutorial.

    Args:
        nb_metadata (Dict[str, Dict[str, str]]): Metadata for the tutorial.

    Returns:
        str: MDX formatted buttons.
    """
    version = get_current_git_tag() or "main"
    github_path = (
        f"{ORGANIZATION}/{PROJECT}/blob/{version}/tutorials/"
        f"{nb_metadata['id']}/{nb_metadata['id']}.ipynb"
    )
    github_url = f"https://github.com/{github_path}"
    colab_url = f"https://colab.research.google.com/github/{github_path}"
    return f'<LinkButtons\n  githubUrl="{github_url}"\n  colabUrl="{colab_url}"\n/>\n\n'


def handle_image_attachments(
    markdown: str,
    attachments: dict[str, dict[str, str]],
) -> str:
    """
    Image attachments are stored in the notebook cell's "attachments" field in base64
    format with their associated mime_type and referenced in the markdown via
    attachment name.

    The pattern we search for in the Markdown is
    `![alt_text](attachment:attachment_name title)` with three groups:

    - group 1 = alt_text (optional)
    - group 2 = attachment_name
    - group 3 = title (optional)

    To represent this in MD we replace the attachment reference with the base64 encoded
    string as `![{alt_text}](data:{mime_type};base64,{img_as_base64})`

    Args:
        markdown (str): The markdown content containing image attachments.
        attachments (Dict[str, Dict[str, str]]): A dictionary of attachments with their
            corresponding MIME types and base64 encoded data.

    Returns:
        str: The markdown content with images converted to base64 format.
    """
    markdown_image_pattern = re.compile(
        r"""!\[([^\]]*)\]\(attachment:(.*?)\s*?(\".*\")?\)"""
    )
    # go through searches in reverse order so that each replacement doesn't affect the
    # start/end indices for the next replacements
    searches = reversed(list(re.finditer(markdown_image_pattern, markdown)))
    for search in searches:
        alt_text, attachment_name, _ = search.groups()
        mime_type, base64 = next(iter(attachments[attachment_name].items()))
        start, end = search.span()
        markdown = (
            markdown[:start]
            + generate_img_base64_md(base64, mime_type, alt_text)
            + markdown[end:]
        )
    return markdown


def handle_image_paths_found_in_markdown(
    markdown: str,
    new_img_dir: Path,
    nb_path: Path,
) -> str:
    """
    Update image paths in the Markdown, and copy the image to the docs location.

    The pattern we search for in the Markdown is
    ``![alt-text](path/to/image.png "title")`` with two groups:

    - group 1 = path/to/image.png
    - group 2 = "title"

    We explicitly exclude matching if the path starts with `attachment:` as this
    indicates that the image is embedded as a base64 attachment not a file path.

    The first group (the path to the image from the original notebook) will be replaced
    with ``assets/img/{name}`` where the name is `image.png` from the example above. The
    original image will also be copied to the new location
    ``{new_img_dir}/assets/img/{name}``, which can be directly read into the MDX file.

    Args:
        markdown (str): Markdown where we look for Markdown flavored images.
        new_img_dir (Path): Path where images are copied to for display in the
            MDX file.
        lib_dir (Path): The location for the repo.
        nb_path (Path): The location for the notebook.

    Returns:
        str: The original Markdown with new paths for images.
    """
    markdown_image_pattern = re.compile(
        r"""!\[[^\]]*\]\((?!attachment:)(.*?)(?=\"|\))(\".*\")?\)"""
    )
    searches = list(re.finditer(markdown_image_pattern, markdown))

    # Return the given Markdown if no images are found.
    if not searches:
        return markdown

    # Process searches in reverse order so that each replacement doesn't affect the
    # start/end indices for the next replacements
    for search in reversed(searches):
        # Find the old image path and replace it with the new one.
        old_path, _ = search.groups()
        # Get the span of the old_path within the full match
        start, end = search.span(1)

        old_path = Path(old_path)
        name = old_path.name.strip()
        new_path = f"assets/img/{name}"

        # Replace the old path with the new path in the markdown
        markdown = markdown[:start] + new_path + markdown[end:]

        # Copy the original image to the new location.
        if old_path.exists():
            # resolves if an absolute path is used
            old_img_path = old_path
        else:
            # fall back to path relative to the notebook
            old_img_path = (nb_path.parent / old_path).resolve()
        new_img_path = str(new_img_dir / name)
        shutil.copy(str(old_img_path), new_img_path)

    return markdown


def transform_style_attributes(markdown: str) -> str:
    """
    Convert HTML style attributes to something React can consume.

    Args:
        markdown (str): Markdown where we look for HTML style attributes.

    Returns:
        str: The original Markdown with new React style attributes.
    """
    # Finds all instances of `style="attr: value; ..."`.
    token = "style="
    pattern = re.compile(f"""{token}["'`]([^"]*)["'`]""")
    found_patterns = re.findall(pattern, markdown)
    if not found_patterns:
        return markdown

    for found_pattern in found_patterns:
        # Step 1: splits "attr: value; ..." to
        #                ["attr: value", ..."].
        step1 = [token.strip() for token in found_pattern.split(";") if token]

        # Step 2: splits ["attr: value", ...] to
        #                [["attr", "value"], ...].
        step2 = [[token.strip() for token in tokens.split(":")] for tokens in step1]

        # Step 3: converts [["attr", "value"], ...] to
        #                  '{"attr": "value", ...}'.
        step3 = json.dumps(dict(step2))

        # Step 4 wraps the JSON object in {}, so we end up with a string of the form;
        #        '{{"attr": "value", ...}}'.
        step4 = f"{{{step3}}}"

        # Step 5 replaces the old style data with the React style data, and clean the
        #        string for inclusion in the final Markdown.
        markdown = markdown.replace(found_pattern, step4)
        markdown = markdown.replace('"{{', "{{").replace('}}"', "}}")
    return markdown


def sanitize_mdx(mdx: str) -> str:
    """
    Sanitize the given MDX string.

    Args:
        mdx (str): MDX string to sanitize.

    Returns:
        str: Sanitized MDX string.
    """
    # Remove some lingering HTML tags that break MDX.
    mdx = mdx.replace("<markdowncell>", "")
    mdx = mdx.replace("<TODO>", "")
    mdx = mdx.replace("<br>", "<br />")
    # Remove any HTML comments from the Markdown. They are fine to keep in the
    # notebooks, but are not really useful in the MDX.
    mdx = re.sub("(<!--.*?-->)", "", mdx, flags=re.DOTALL)
    # "\" Escape braces to make the text MDX compatible.
    mdx = re.sub("([^\\\\])([{}])", "\\g<1>\\\\\\g<2>", mdx)

    # -- KaTeX --
    # Make sure $$ symbols are not escaped and include line breaks.
    mdx = re.sub(
        "\\\\?\\$\\\\?\\$((?:.|\n)*?)\\\\?\\$\\\\?\\$", "\n$$\n\\g<1>\n$$\n", mdx
    )
    # Escaping braces causes issues in math blocks, unescape them.
    mdx = re.sub(
        "\\$?\\$(.|\n)*?\\$\\$?",
        lambda match: match[0].replace("\\{", "{").replace("\\}", "}"),
        mdx,
    )

    return mdx


def get_source(cell: NotebookNode) -> str:
    """
    Extract the source code from a Jupyter notebook cell. The source is
    characteristically multi-line strings but may be stored as a list of strings, in
    which case we join them together.
    https://ipython.readthedocs.io/en/3.x/notebook/nbformat.html

    Args:
        cell (NotebookNode): A Jupyter notebook cell object.

    Returns:
        str: The source code as a string.
    """
    source = cell.get("source", "")
    if isinstance(source, list):
        return "".join(source)
    return source


def handle_markdown_cell(
    cell: NotebookNode,
    new_img_dir: Path,
    nb_path: Path,
) -> str:
    """
    Handle the given Jupyter Markdown cell and convert it to MDX.

    Args:
        cell (NotebookNode): Jupyter Markdown cell object.
        new_img_dir (Path): Path where images are copied to for display in the
            Markdown cell.
        lib_dir (Path): The location for the library.
        nb_path (Path): The location for the notebook.

    Returns:
        str: Transformed Markdown object suitable for inclusion in MDX.
    """
    markdown = get_source(cell)

    # Handle the different ways images are included in the Markdown.
    markdown = handle_image_paths_found_in_markdown(markdown, new_img_dir, nb_path)
    markdown = handle_image_attachments(markdown, cell.get("attachments", {}))

    markdown = sanitize_mdx(markdown)
    mdx = mdformat.text(markdown, options={"wrap": 88}, extensions={"myst"})

    # We will attempt to handle inline style attributes written in HTML by converting
    # them to something React can consume. This has to be done after the
    # mdformat.text() step since mdformat complains about the converted styles.
    mdx = transform_style_attributes(mdx)

    return f"{mdx}\n"


def handle_cell_input(cell: NotebookNode, language: str) -> str:
    """
    Create a Markdown cell block using the given cell source, and the language.

    The given language will determine cell input syntax styles. Docusaurus uses Prism as
    the syntax highlighter, https://prismjs.com. See the Docusaurus documentation for
    more information on code blocks
    https://docusaurus.io/docs/markdown-features/code-blocks.

    Args:
        cell (NotebookNode): A notebook cell.
        language (str): Language specifier for syntax highlighting.

    Returns:
        str: Code block formatted Markdown string.
    """
    cell_source = get_source(cell)
    return f"```{language}\n{cell_source}\n```\n\n"


def generate_img_base64_md(
    img_as_base64: int | str | NotebookNode,
    mime_type: int | str | NotebookNode,
    alt_text: str = "",
) -> str:
    """
    Generate a markdown image tag from a base64 encoded image.

    Args:
        img_as_base64 (int | str | NotebookNode): The base64 encoded image data.
        mime_type (int | str | NotebookNode): The MIME type of the image.
        alt_text (str, optional): The alternative text for the image. Defaults to an
            empty string.

    Returns:
        str: A markdown formatted image tag.
    """
    return f"![{alt_text}](data:{mime_type};base64,{img_as_base64})"


def handle_image(
    values: list[dict[str, int | str | NotebookNode]],
) -> list[tuple[int, str]]:
    """
    Convert embedded images to string MDX can consume.

    Args:
        values (List[Dict[str, Union[int, str, NotebookNode]]]): Bokeh tagged cell
            outputs.

    Returns:
        List[Tuple[int, str]]: A list of tuples, where the first entry in the tuple is
            the index where the output occurred from the cell, and the second entry of
            the tuple is the MDX formatted string.
    """
    output = []
    for value in values:
        index = value["index"]
        mime_type = value["mime_type"]
        img = value["data"]
        output.append((index, f"{generate_img_base64_md(img, mime_type)}\n\n"))
    return output


def handle_markdown(
    values: list[dict[str, int | str | NotebookNode]],
) -> list[tuple[int, str]]:
    """
    Convert and format Markdown for MDX.

    Args:
        values (List[Dict[str, Union[int, str, NotebookNode]]]): Bokeh tagged cell
            outputs.

    Returns:
        List[Tuple[int, str]]: A list of tuples, where the first entry in the tuple is
            the index where the output occurred from the cell, and the second entry of
            the tuple is the MDX formatted string.
    """
    output = []
    for value in values:
        index = int(value["index"])
        markdown = str(value["data"])
        markdown = mdformat.text(markdown, options={"wrap": 88}, extensions={"myst"})
        output.append((index, f"{markdown}\n\n"))
    return output


def handle_pandas(
    values: list[dict[str, int | str | NotebookNode]],
) -> list[tuple[int, str]]:
    """
    Handle how to display pandas DataFrames.

    There is a scoped style tag in the DataFrame output that uses the class name
    `dataframe` to style the output. We will use this token to determine if a pandas
    DataFrame is being displayed.

    Args:
        values (List[Dict[str, Union[int, str, NotebookNode]]]): Bokeh tagged cell
            outputs.

    Returns:
        List[Tuple[int, str]]: A list of tuples, where the first entry in the tuple is
            the index where the output occurred from the cell, and the second entry of
            the tuple is the MDX formatted string.
    """
    output = []
    for value in values:
        index = int(value["index"])
        data = str(value["data"])
        if 'class="dataframe"' not in data:
            # pd.read_html() raises an error if there's no dataframe.
            continue
        df = pd.read_html(io.StringIO(data), flavor="lxml")
        if len(df) == 0:
            continue
        # NOTE: The return is a list of dataframes and we only care about the first
        #       one.
        md_df = df[0]
        for column in md_df.columns:
            if column.startswith("Unnamed"):
                md_df.rename(columns={column: ""}, inplace=True)
        # Remove the index if it is just a range, and output to markdown.
        mdx = ""
        if isinstance(md_df.index, pd.RangeIndex):
            mdx = md_df.to_markdown(index=False)
        elif not isinstance(md_df.index, pd.RangeIndex):
            mdx = md_df.to_markdown()
        output.append((index, f"\n{sanitize_mdx(mdx)}\n\n"))
    return output


def handle_plain(
    values: list[dict[str, int | str | NotebookNode]],
) -> list[tuple[int, str]]:
    """
    Handle how to plain cell output should be displayed in MDX.

    Stdout streams are chunked during execution, we merge adjacent streams here into
    single cell output blocks.

    Args:
        values (List[Dict[str, Union[int, str, NotebookNode]]]): Bokeh tagged cell
            outputs.

    Returns:
        List[Tuple[int, str]]: A list of tuples, where the first entry in the tuple is
            the index where the output occurred from the cell, and the second entry of
            the tuple is the MDX formatted string.
    """
    output = []
    adjacent_outputs = []
    previous_index = -1

    def append_to_output() -> None:
        if not adjacent_outputs:
            return
        adjacent_outputs_str = "\n".join(adjacent_outputs)
        output.append(
            (
                previous_index,
                "\n".join(
                    [
                        "<CellOutput>",
                        "{",
                        f"`{adjacent_outputs_str}`",
                        "}",
                        "</CellOutput>\n\n",
                    ]
                ),
            ),
        )

    for value in values:
        index = int(value["index"])
        data = str(value["data"])
        data = [line.strip() for line in data.splitlines() if line]
        data = [datum for datum in data if datum]
        if data:
            data = "\n".join([line for line in str(value["data"]).splitlines() if line])
            # Remove backticks to make the text MDX compatible.
            data = data.replace("`", "")
            if previous_index in [-1, index - 1]:
                # store in cache until we reach nonconsecutive index
                adjacent_outputs.append(data)
            else:
                # flush cache to output and start a new one
                append_to_output()
                adjacent_outputs = [data]
            previous_index = index
    # flush the remaining cache to output
    append_to_output()
    return output


def handle_plotly(
    values: list[dict[str, int | str | NotebookNode]],
    plot_data_folder: Path,
) -> list[tuple[int, str]]:
    """
    Convert Plotly outputs to MDX.

    Args:
        values (List[Dict[str, Union[int, str, NotebookNode]]]): Bokeh tagged cell
            outputs.
        plot_data_folder (Path): Path to the folder where plot data should be
            stored.

    Returns:
        List[Tuple[int, str]]: A list of tuples, where the first entry in the tuple is
            the index where the output occurred from the cell, and the second entry of
            the tuple is the MDX formatted string.
    """
    output = []
    for value in values:
        index = value["index"]
        data = value["data"]
        file_name = str(uuid.uuid4())
        file_path = plot_data_folder / f"{file_name}.json"
        path_to_data = f"./assets/plot_data/{file_name}.json"
        output.append(
            (index, f"<PlotlyFigure data={{require('{path_to_data}')}} />\n\n"),
        )
        with file_path.open("w") as f:
            json.dump(data, f, indent=2)
    return output


def handle_tqdm(
    values: list[dict[str, int | str | NotebookNode]],
) -> list[tuple[int, str]]:
    """
    Handle the output of tqdm.

    tqdm will be displayed as separate CellOutput React components if we do not
    aggregate them all into a single CellOutput object, which is what this method does.

    Args:
        values (List[Dict[str, Union[int, str, NotebookNode]]]): Bokeh tagged cell
            outputs.

    Returns:
        List[Tuple[int, str]]: A list of tuples, where the first entry in the tuple is
            the index where the output occurred from the cell, and the second entry of
            the tuple is the MDX formatted string.
    """
    output = sorted(values, key=lambda item: item["index"])
    index = int(output[0]["index"])
    md = "\n".join([str(item["data"]) for item in output if item["data"]])
    return [(index, f"<CellOutput>\n{{\n  `{md}`\n}}\n</CellOutput>\n\n")]


CELL_OUTPUTS_TO_PROCESS = dict[
    str,
    list[dict[str, Union[int, str, NotebookNode]]],
]


def aggregate_mdx(
    cell_outputs_to_process: CELL_OUTPUTS_TO_PROCESS,
    plot_data_folder: Path,
) -> str:
    """
    Aggregate the `cell_outputs_to_process` into MDX.

    Args:
        cell_outputs_to_process (CELL_OUTPUTS_TO_PROCESS): A dictionary of cell outputs
            that need further processing.
        plot_data_folder (Path): Path to where plot data should be stored for the
            tutorial.

    Returns:
        str: MDX formatted string.
    """
    processed_mdx = []
    for key, values in cell_outputs_to_process.items():
        if not values:
            continue
        if key == "image":
            processed_mdx.extend(handle_image(values))
        if key == "markdown":
            processed_mdx.extend(handle_markdown(values))
        if key == "pandas":
            processed_mdx.extend(handle_pandas(values))
        if key == "plain":
            processed_mdx.extend(handle_plain(values))
        if key == "plotly":
            processed_mdx.extend(handle_plotly(values, plot_data_folder))
        if key == "tqdm":
            processed_mdx.extend(handle_tqdm(values))

    # Ensure the same ordering of the MDX happens as was found in the original cell
    # output.
    processed_mdx = sorted(processed_mdx, key=lambda item: item[0])
    mdx = "\n".join([item[1] for item in processed_mdx])
    return mdx


def prioritize_dtypes(
    cell_outputs: list[NotebookNode],
) -> tuple[list[list[str]], list[bool]]:
    """
    Prioritize cell output data types.

    Args:
        cell_outputs (List[NotebookNode]): A list of cell outputs.

    Returns:
        Tuple[List[List[str]], List[bool]]: Return two items in the tuple; the first is
            a list of prioritized data types and the second is a list boolean values
            associated with the cell output having Plotly information in it or not.
    """
    cell_output_dtypes = [
        (
            list(cell_output["data"].keys())
            if "data" in cell_output
            else [cell_output["output_type"]]
        )
        for cell_output in cell_outputs
    ]
    prioritized_cell_output_dtypes = [
        sorted(
            set(dtypes).intersection(set(priorities)),
            key=lambda dtype: priorities.index(dtype),
        )
        for dtypes in cell_output_dtypes
    ]
    prioritized_cell_output_dtypes = [
        [str(item) for item in items] for items in prioritized_cell_output_dtypes
    ]
    plotly_flags = [
        any(["plotly" in output for output in outputs])
        for outputs in cell_output_dtypes
    ]
    return prioritized_cell_output_dtypes, plotly_flags


def aggregate_images_and_plotly(
    prioritized_data_dtype: str,
    cell_output: NotebookNode,
    data: NotebookNode,
    plotly_flags: list[bool],
    cell_outputs_to_process: CELL_OUTPUTS_TO_PROCESS,
    i: int,
) -> None:
    """
    Aggregates images or Plotly cell outputs into an appropriate bucket.

    Args:
        prioritized_data_dtype (str): The prioritized cell output data type.
        cell_output (NotebookNode): The actual cell output from the notebook.
        data (NotebookNode): The data of the cell output.
        plotly_flags (List[bool]): True if a Plotly plot was found in the cell outputs
            else False.
        cell_outputs_to_process (CELL_OUTPUTS_TO_PROCESS): Dictionary containing
            aggregated cell output objects.
        i (int): Index for the cell output in the list of cell output objects.

    Returns:
        None: Does not return anything, instead adds values to the
            cell_outputs_to_process if applicable.
    """
    if not plotly_flags[i]:
        cell_outputs_to_process["image"].append(
            {"index": i, "data": data, "mime_type": prioritized_data_dtype},
        )
    # Plotly outputs a static image, but we can use the JSON in the cell
    # output to create interactive plots using a React component.
    if plotly_flags[i]:
        data = cell_output["data"]["application/vnd.plotly.v1+json"]
        cell_outputs_to_process["plotly"].append({"index": i, "data": data})


def aggregate_plain_output(
    prioritized_data_dtype: str,
    cell_output: NotebookNode,
    data: NotebookNode,
    cell_outputs_to_process: CELL_OUTPUTS_TO_PROCESS,
    i: int,
) -> None:
    """
    Aggregate plain text cell outputs together.

    Args:
        prioritized_data_dtype (str): The prioritized cell output data type.
        cell_output (NotebookNode): The actual cell output from the notebook.
        data (NotebookNode): The data of the cell output.
        cell_outputs_to_process (CELL_OUTPUTS_TO_PROCESS): Dictionary containing
            aggregated cell output objects.
        i (int): Index for the cell output in the list of cell output objects.

    Returns:
        None: Does not return anything, instead adds values to the
            cell_outputs_to_process if applicable.
    """
    # Ignore error outputs.
    if "name" in cell_output and cell_output["name"] == "stderr":
        pass
    # Ignore matplotlib legend text output.
    if prioritized_data_dtype == "text/plain" and "matplotlib" in data:
        pass
    cell_outputs_to_process["plain"].append({"index": i, "data": data})


def aggregate_output_types(cell_outputs: list[NotebookNode]) -> CELL_OUTPUTS_TO_PROCESS:
    """
    Aggregate cell outputs into a dictionary for further processing.

    Args:
        cell_outputs (List[NotebookNode]): List of cell outputs.

    Returns:
        CELL_OUTPUTS_TO_PROCESS: Dictionary containing aggregated cell output objects.
    """
    # We will use the below cell output data types for prioritizing the output shown in
    # the MDX file.
    prioritized_cell_output_dtypes, plotly_flags = prioritize_dtypes(cell_outputs)

    cell_outputs_to_process = {
        "image": [],
        "markdown": [],
        "pandas": [],
        "plain": [],
        "plotly": [],
        "tqdm": [],
    }
    for i, cell_output in enumerate(cell_outputs):
        prioritized_data_dtype = prioritized_cell_output_dtypes[i][0]

        # If there is no `data` key in the cell_output, then it may be an error that
        # needs to be handled. Even if it is not an error, the data is stored in a
        # different key if no `data` key is found.
        data = (
            cell_output["data"][prioritized_data_dtype]
            if "data" in cell_output
            else cell_output["text"]
            if "text" in cell_output
            else cell_output["evalue"]
        )
        image_check = (
            prioritized_data_dtype.startswith("image")
            or "plotly" in prioritized_data_dtype
        )
        if image_check:
            aggregate_images_and_plotly(
                prioritized_data_dtype,
                cell_output,
                data,
                plotly_flags,
                cell_outputs_to_process,
                i,
            )
        plain_check = prioritized_data_dtype in ["text/plain", "stream", "error"]
        if plain_check:
            aggregate_plain_output(
                prioritized_data_dtype,
                cell_output,
                data,
                cell_outputs_to_process,
                i,
            )
        if prioritized_data_dtype == "text/markdown":
            cell_outputs_to_process["markdown"].append({"index": i, "data": data})
        if "dataframe" in data:
            cell_outputs_to_process["pandas"].append({"index": i, "data": data})
        if prioritized_data_dtype == "application/vnd.jupyter.widget-view+json":
            data = cell_output["data"]["text/plain"]
            cell_outputs_to_process["tqdm"].append({"index": i, "data": data})

    return cell_outputs_to_process


def handle_cell_outputs(cell: NotebookNode, plot_data_folder: Path) -> str:
    """
    Handle cell outputs and convert to MDX.

    Args:
        cell (NotebookNode): The cell where the outputs need converting.
        plot_data_folder (Path): Path to the folder where plot data should be
            stored.

    Returns:
        str: MDX formatted cell output.
    """
    mdx = ""

    # Return an empty string if there are no actual cell outputs.
    cell_outputs = cell.get("outputs", [])
    if not cell_outputs:
        return mdx

    # We will loop over all cell outputs and bucket them into the appropriate key in the
    # dictionary below for further processing. Doing it in this way helps aggregate like
    # outputs together e.g. tqdm outputs.
    cell_outputs_to_process = aggregate_output_types(cell_outputs)

    # Now we process all aggregated cell outputs into a single output for the type.
    md = aggregate_mdx(cell_outputs_to_process, plot_data_folder)
    return md


def handle_code_cell(cell: NotebookNode, plot_data_folder: Path) -> str:
    """
    Handle code cells in Jupyter notebooks and convert them to MDX.

    Args:
        cell (NotebookNode): A Jupyter notebook cell that contains code.
        plot_data_folder (Path): Path to the folder where plot data should be
            stored.

    Returns:
        str: MDX formatted code cell.
    """
    cell_input_mdx = handle_cell_input(cell, "python")
    cell_output_mdx = handle_cell_outputs(cell, plot_data_folder)
    return cell_input_mdx + cell_output_mdx


def transform_notebook(path: Path, nb_metadata: object) -> str:
    """
    Transform a notebook located at the given path into MDX.

    Args:
        path (Path): Path to the Jupyter notebook tutorial.

    Returns:
        str: MDX formatted string.
    """
    filename, assets_folder = create_folders(path)
    img_folder = assets_folder / "img"
    plot_data_folder = assets_folder / "plot_data"
    save_folder = assets_folder.joinpath("..").resolve()
    nb = load_notebook(path)
    mdx = ""
    mdx += create_frontmatter(path, nb_metadata)
    mdx += create_imports()
    mdx += create_buttons(nb_metadata)
    for cell in nb["cells"]:
        cell_type = cell["cell_type"]

        # Handle a Markdown cell.
        if cell_type == "markdown":
            mdx += handle_markdown_cell(cell, img_folder, path)

        # Handle a code cell.
        if cell_type == "code":
            mdx += handle_code_cell(cell, plot_data_folder)

    # Write the MDX file to disk.
    save_path = save_folder / "index.mdx"
    with save_path.open("w") as f:
        f.write(mdx)

    # Return the string for debugging purposes.
    return mdx


def clean_up_directories() -> None:
    """
    Delete output from previously converted notebooks. Particularly useful for
    removing plot data since those filenames are randomly generated and won't
    be replaced with future runs.

    Returns:
        None: Does not return anything.
    """
    if TUTORIALS_DIR.exists():
        # We intentionally leave the static `index.mdx` file in place since that is not
        # autogenerated.
        for item in os.scandir(TUTORIALS_DIR):
            if item.is_dir():
                shutil.rmtree(item.path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert tutorial notebooks into mdx files."
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        help="Delete output from previously converted notebooks.",
    )
    args = parser.parse_args()

    tutorials_metadata = load_nb_metadata()
    print("--------------------------------------------")
    print("Converting tutorial notebooks into mdx files")
    print("--------------------------------------------")
    if args.clean:
        clean_up_directories()
    for metadata in tutorials_metadata:
        path = (
            LIB_DIR / "tutorials" / metadata["id"] / (metadata["id"] + ".ipynb")
        ).resolve()
        print(f"{path.stem}")
        mdx = transform_notebook(path, metadata)
    print("")
